#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading FLAGSTATION module.')

def _run_parser(soltab, parser, step):
    mode = parser.getstr( step, 'mode') # no default
    maxFlaggedFraction = parser.getfloat( step, 'maxFlaggedFraction', 0.5)
    nSigma = parser.getfloat( step, 'nSigma', 5.0)
    maxStddev = parser.getfloat( step, 'maxStddev', -1.0)
    ampRange = parser.getarrayfloat( step, 'ampRange', [0.,0.] )
    telescope = parser.getstr( step, 'telescope', 'lofar')
    skipInternational = parser.getbool( step, 'skipInternational', False)
    refAnt = parser.getstr( step, 'refAnt', '')
    soltabExport = parser.getstr( step, 'soltabExport', '' )
    ncpu = parser.getint( '_global', 'ncpu', 0)

    parser.checkSpelling( step, soltab, ['mode', 'maxFlaggedFraction', 'nSigma', 'maxStddev', 'ampRange', 'telescope', 'skipInternational', 'refAnt', 'soltabExport'])
    return run( soltab, mode, maxFlaggedFraction, nSigma, maxStddev, ampRange, telescope, skipInternational, refAnt, soltabExport, ncpu )


def _flag_resid(vals, weights, soltype, nSigma, maxFlaggedFraction, maxStddev, ants, s, outQueue):
    """
    Flags bad residuals relative to mean by setting the corresponding weights to 0.0

    Parameters
    ----------
    vals : array
        Array of values as [time, ant, freq, pol]

    weights : array
        Array of weights as [time, ant, freq, pol]

    soltype : str
        Type of solutions: phase or amplitude

    nSigma : float
        Number of sigma for flagging. vals outside of nSigma*stddev are flagged

    maxFlaggedFraction : float
        Maximum allowable fraction of flagged frequencies. Stations with higher fractions
        will be completely flagged

    maxStddev : float
        Maximum allowable standard deviation

    ants : list
        List of station names

    s : int
        Station index

    Returns
    -------
    indx, weights : int, array
        Station index, modified weights array
    """
    # Skip fully flagged stations
    if np.all(weights == 0.0):
        outQueue.put([s, weights])
        return

    # Iterate over polarizations
    npols = vals.shape[2] # number of polarizations
    for pol in range(npols):
        # Check flags
        weights_orig = weights[:, :, pol]
        if soltype=='phase':
            bad_sols = np.where(np.isnan(vals[:, :, pol]))
        else:
            bad_sols = np.where(np.logical_or(np.isnan(vals[:, :, pol]), vals[:, :, pol] <= 0.0))
        weights_orig[bad_sols] = 0.0
        if np.all(weights_orig == 0.0):
            # Skip fully flagged polarizations
            continue
        flagged = np.where(weights_orig == 0.0)
        unflagged = np.where(weights_orig != 0.0)

        if soltype == 'amplitude':
            # Take the log
            vals[:, :, pol] = np.log10(vals[:, :, pol])

        # Remove mean (to avoid wraps near +/- pi) and set flagged points to 0
        if soltype=='phase':
            mean = np.angle( np.nansum( weights_orig.flatten() * np.exp(1j*vals[:, :, pol].flatten()) ) / ( vals[:, :, pol].flatten().size * sum(weights_orig.flatten()) ) )
        else:
            mean = np.nansum( weights_orig.flatten() * vals[:, :, pol].flatten() ) / ( vals[:, :, pol].flatten().size * sum(weights_orig.flatten()) )
        vals_flagged = vals[:, :, pol]
        if soltype=='phase':
            # Remove the mean to avoid wrapping issues near +/- pi
            vals_flagged = normalize_phase(vals_flagged - mean)
        vals_flagged[flagged] = 0.0

        # Iteratively fit and flag
        nsols_unflagged = len(vals_flagged[unflagged])
        maxiter = 5
        niter = 0
        nflag = 0
        nflag_prev = -1
        weights_copy = weights_orig.copy()
        while nflag != nflag_prev and niter < maxiter:
            stdev_all = np.sqrt(np.average(vals_flagged**2, weights=weights_copy))
            stdev = min(maxStddev, stdev_all)
            bad = np.where(np.abs(vals_flagged) > nSigma*stdev)
            nflag = len(bad[0])
            if nflag == 0 or nflag == nsols_unflagged:
                break
            if niter > 0:
                nflag_prev = nflag
            weights_copy = weights_orig.copy()  # reset flags to original ones
            weights_copy[bad] = 0
            niter += 1

        # Check whether station is bad (high flagged fraction). If
        # so, flag all frequencies and polarizations
        if float(len(bad[0]))/float(nsols_unflagged) > maxFlaggedFraction:
            # Station has high fraction of initially unflagged solutions that are now flagged
            logging.info('Flagged {0} (pol {1}) due to high flagged fraction '
                         '({2:.2f})'.format(ants[s], pol, float(len(bad[0]))/float(nsols_unflagged)))
            weights[:, :, pol] = 0.0
        else:
            # Station is OK, flag bad points only
            nflagged_orig = len(np.where(weights_orig == 0.0)[0])
            nflagged_new = len(np.where(weights_copy == 0.0)[0])
            weights[:, :, pol] = weights_copy
            prcnt = float(nflagged_new - nflagged_orig) / float(np.product(weights_orig.shape)) * 100.0
            logging.info('Flagged {0:.1f}% of solutions for {1} (pol {2})'.format(prcnt, ants[s], pol))

    outQueue.put([s, weights])


def _flag_bandpass(freqs, amps, weights, telescope, nSigma, ampRange, maxFlaggedFraction, maxStddev,
                   plot, ants, s, outQueue):
    """
    Flags bad amplitude solutions relative to median bandpass (in log space) by setting
    the corresponding weights to 0.0

    Note: A median over the time axis is done before flagging, so the flags are not time-
    dependent

    Parameters
    ----------
    freqs : array
        Array of frequencies

    amps : array
        Array of amplitudes as [time, ant, freq, pol]

    weights : array
        Array of weights as [time, ant, freq, pol]

    telescope : str, optional
        Specifies the telescope for the bandpass model

    nSigma : float
        Number of sigma for flagging. Amplitudes outside of nSigma*stddev are flagged

    maxFlaggedFraction : float
        Maximum allowable fraction of flagged frequencies. Stations with higher fractions
        will be completely flagged

    maxStddev : float
        Maximum allowable standard deviation

    plot : bool
        If True, the bandpass with flags and best-fit line is plotted for each station

    ants : list
        List of station names

    s : int
        Station index

    ampRange : array
        2-element array of the median amplitude level to be acceptable, ampRange[0]: lower limit, ampRange[1]: upper limit

    Returns
    -------
    indx, weights : int, array
        Station index, modified weights array
    """
    def _B(x, k, i, t, extrap, invert):
        if k == 0:
            if extrap:
                if invert:
                    return -1.0
                else:
                    return 1.0
            else:
                return 1.0 if t[i] <= x < t[i+1] else 0.0
        if t[i+k] == t[i]:
            c1 = 0.0
        else:
            c1 = (x - t[i])/(t[i+k] - t[i]) * _B(x, k-1, i, t, extrap, invert)
        if t[i+k+1] == t[i+1]:
            c2 = 0.0
        else:
            c2 = (t[i+k+1] - x)/(t[i+k+1] - t[i+1]) * _B(x, k-1, i+1, t, extrap, invert)
        return c1 + c2

    def _bspline(x, t, c, k):
        n = len(t) - k - 1
        assert (n >= k+1) and (len(c) >= n)
        invert = False
        extrap = [False] * n
        if x >= t[n]:
            extrap[-1] = True
        elif x < t[k]:
            extrap[0] = True
            invert = False
        return sum(c[i] * _B(x, k, i, t, e, invert) for i, e in zip(list(range(n)), extrap))

    def _bandpass_LBA(freq, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13, c14, c15):
        """
        Defines the functional form of the LBA bandpass in terms of splines of degree 3

        The spline fit was done using LSQUnivariateSpline() on the median bandpass between
        10 MHz and 78 MHz. The knots were set by hand to acheive a good fit with a
        minimum number of parameters.

        Parameters
        ----------
        freq : array
            Array of frequencies

        c1-c13 : float
            Spline coefficients

        Returns
        -------
        bandpass : list
            List of bandpass values as function of frequency
        """
        knots = np.array([10850524.0, 10850524.0, 10850524.0, 10850524.0,
                          20000000.0, 30000000.0, 40000000.0, 50000000.0,
                          55000000.0, 56000000.0, 60000000.0, 62000000.0,
                          63000000.0, 64000000.0, 70000000.0, 77610779.0,
                          77610779.0, 77610779.0, 77610779.0])
        coeffs = np.array([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13, c14, c15])
        return [_bspline(f, knots, coeffs, 3) for f in freq]

    def _bandpass_HBA_low(freq, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10):
        """
        Defines the functional form of the HBA-low bandpass in terms of splines of degree
        3

        The spline fit was done using LSQUnivariateSpline() on the median bandpass between
        120 MHz and 188 MHz. The knots were set by hand to acheive a good fit with a
        minimum number of parameters.

        Parameters
        ----------
        freq : array
            Array of frequencies

        c1-c10 : float
            Spline coefficients

        Returns
        -------
        bandpass : list
            List of bandpass values as function of frequency
        """
        knots = np.array([1.15e+08, 1.15e+08, 1.15e+08, 1.15e+08,
                          1.30e+08, 1.38e+08, 1.48e+08, 1.60e+08,
                          1.68e+08, 1.78e+08, 1.90e+08, 1.90e+08,
                          1.9e+08, 1.9e+08])
        coeffs = np.array([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10])
        return [_bspline(f, knots, coeffs, 3) for f in freq]

    def _fit_bandpass(freq, logamp, sigma, band, do_fit=True):
        """
        Fits amplitudes with one of the bandpass functions

        The initial coefficients were determined from a LSQUnivariateSpline() fit on the
        median bandpass of the appropriate band. The allowable fitting ranges were set by
        hand through testing on a number of observations (to allow the bandpass function
        to adjust for the differences between stations but not to fit to RFI, etc.).

        Parameters
        ----------
        freq : array
            Array of frequencies

        amps : array
            Array of log10(amplitudes)

        sigma : array
            Array of sigma (1/weights**2)

        band : str
            Band name ('hba_low', etc.)

        do_fit : bool, optional
            If True, the fitting is done. If False, the unmodified model bandpass is
            returned

        Returns
        -------
        fit_parms, bandpass : list, list
            List of best-fit parameters, List of bandpass values as function of frequency
        """
        from scipy.optimize import curve_fit

        if band.lower() == 'hba_low':
            bandpass_function = _bandpass_HBA_low
            init_coeffs = np.array([-0.01460369, 0.05062699, 0.02827004, 0.03738518,
                                    -0.05729109, 0.02303295, -0.03550487, -0.0803113,
                                    -0.2394929, -0.358301])
            bounds_deltas_lower = [0.06, 0.05, 0.04, 0.04, 0.04, 0.04, 0.1, 0.1, 0.2, 0.5]
            bounds_deltas_upper = [0.06, 0.1, 0.1, 0.1, 0.04, 0.04, 0.04, 0.04, 0.05, 0.06]
        elif band.lower() == 'lba':
            bandpass_function = _bandpass_LBA
            init_coeffs = np.array([-0.19996913, -0.10604762, -0.29063847, -0.17248618,  0.05588799,
                                    0.21518069,  0.46090289,  0.59503671,  0.6278482 ,  0.42936404,
                                    0.43576633,  0.29100915, -0.00448301, -0.16566164, -0.3667639 ])
            bounds_deltas_lower = [0.1, 0.2, 0.25, 0.2, 0.05, 0.05, 0.05, 0.1, 0.1, 0.16, 0.2, 0.15,
                                   0.15, 0.25, 0.3]
            bounds_deltas_upper = [0.1, 0.3, 0.4, 0.3, 0.15, 0.05, 0.05, 0.05, 0.08, 0.05, 0.08, 0.15,
                                   0.15, 0.25, 0.35]
        else:
            logging.error('The "{}" band is not supported'.format(band))
            return None, None

        if do_fit:
            lower = [c - b for c, b in zip(init_coeffs, bounds_deltas_lower)]
            upper = [c + b for c, b in zip(init_coeffs, bounds_deltas_upper)]
            param_bounds = (lower, upper)
            try:
                popt, pcov = curve_fit(bandpass_function, freq, logamp, sigma=sigma,
                                       bounds=param_bounds, method='dogbox', ftol=1e-3,
                                       xtol=1e-3, gtol=1e-3)
                return popt, bandpass_function(freq, *tuple(popt))
            except RuntimeError:
                logging.error('Fitting failed.' )
                return None, bandpass_function(freq, *tuple(init_coeffs))
        else:
            return None, bandpass_function(freq, *tuple(init_coeffs))

    # Check that telescope and band is supported. Skip flagging if not
    if telescope.lower() == 'lofar':
        # Determine which band we're in
        if np.median(freqs) < 180e6 and np.median(freqs) > 110e6:
            band = 'hba_low'
        elif np.median(freqs) < 90e6:
            band = 'lba'
        else:
            logging.warning('The median frequency of {} Hz is outside of the currently supported LOFAR bands '
                            '(LBA and HBA-low). Flagging will be skipped'.format(np.median(freqs)))
            outQueue.put([s, weights])
            return
    else:
        logging.warning("Only telescope = 'lofar' is currently supported for bandpass mode. "
                        "Flagging will be skipped")
        outQueue.put([s, weights])
        return

    # Skip fully flagged stations
    if np.all(weights == 0.0):
        outQueue.put([s, weights])
        return

    # Build arrays for fitting
    flagged = np.where(np.logical_or(weights == 0.0, np.isnan(amps)))
    amps_flagged = amps.copy()
    amps_flagged[flagged] = np.nan
    sigma = weights.copy()
    sigma[flagged] = 1.0
    sigma = np.sqrt(1.0 / sigma)
    sigma[flagged] = 1e8

    # Set range of allowed values for the median
    if ampRange is None or ampRange == [0.0, 0.0]:
        # Use sensible values depending on correlator
        if np.nanmedian(amps_flagged) > 1.0:
            # new correlator
            ampRange = [50.0, 325.0]
        else:
            # old correlator
            ampRange = [0.0004, 0.0018]
    median_min = ampRange[0]
    median_max = ampRange[-1]

    # Iterate over polarizations
    npols = amps.shape[2]
    for pol in range(npols):
        # Skip fully flagged polarizations
        if np.all(weights[:, :, pol] == 0.0):
            continue

        # Take median over time and divide out the median offset
        with np.warnings.catch_warnings():
            # Filter NaN warnings -- we deal with NaNs below
            np.warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
            amps_div = np.nanmedian(amps_flagged[:, :, pol], axis=0)
            median_val = np.nanmedian(amps_div)
        amps_div /= median_val
        sigma_div = np.median(sigma[:, :, pol], axis=0)
        sigma_orig = sigma_div.copy()
        unflagged = np.where(~np.isnan(amps_div))
        nsols_unflagged = len(unflagged[0])
        median_flagged = np.where(np.isnan(amps_div))
        amps_div[median_flagged] = 1.0
        sigma_div[median_flagged] = 1e8
        median_flagged = np.where(amps_div <= 0.0)
        amps_div[median_flagged] = 1.0
        sigma_div[median_flagged] = 1e8

        # Before doing the fitting, renormalize and flag any solutions that deviate from
        # the model bandpass by a large factor to avoid biasing the first fit
        _, bp_sp = _fit_bandpass(freqs, np.log10(amps_div), sigma_div, band, do_fit=False)
        normval = np.median(np.log10(amps_div) - bp_sp) # value to normalize model to data
        amps_div /= 10**normval
        bad = np.where(np.abs(np.array(bp_sp) - np.log10(amps_div)) > 0.2)
        sigma_div[bad] = 1e8
        if np.all(sigma_div > 1e7):
            logging.info('Flagged {0} (pol {1}) due to poor match to '
                         'baseline bandpass model'.format(ants[s], pol))
            weights[:, :, pol] = 0.0
            outQueue.put([s, weights])
            return

        # Iteratively fit and flag
        maxiter = 5
        niter = 0
        nflag = 0
        nflag_prev = -1
        while nflag != nflag_prev and niter < maxiter:
            p, bp_sp = _fit_bandpass(freqs, np.log10(amps_div), sigma_div, band)
            stdev_all = np.sqrt(np.average((bp_sp-np.log10(amps_div))**2, weights=(1/sigma_div)**2))
            stdev = min(maxStddev, stdev_all)
            bad = np.where(np.abs(bp_sp - np.log10(amps_div)) > nSigma*stdev)
            nflag = len(bad[0])
            if nflag == 0 or nflag == nsols_unflagged:
                break
            if niter > 0:
                nflag_prev = nflag
            sigma_div = sigma_orig.copy()  # reset flags to original ones
            sigma_div[bad] = 1e8
            niter += 1

        if plot:
            import matplotlib.pyplot as plt
            plt.plot(freqs, bp_sp, 'g-', lw=3)
            plt.plot(freqs, np.log10(amps_div), 'o', c='g')
            plt.plot(freqs[bad], np.log10(amps_div)[bad], 'o', c='r')
            plt.show()

        # Check whether entire station is bad (high stdev or high flagged fraction). If
        # so, flag all frequencies and polarizations
        if stdev_all > nSigma*maxStddev:
            # Station has high stddev relative to median bandpass
            logging.info('Flagged {0} (pol {1}) due to high stddev '
                         '({2})'.format(ants[s], pol, stdev_all))
            weights[:, :, pol] = 0.0
        elif float(len(bad[0]))/float(nsols_unflagged) > maxFlaggedFraction:
            # Station has high fraction of initially unflagged solutions that are now flagged
            logging.info('Flagged {0} (pol {1}) due to high flagged fraction '
                         '({2:.2f})'.format(ants[s], pol, float(len(bad[0]))/float(nsols_unflagged)))
            weights[:, :, pol] = 0.0
        else:
            flagged = np.where(sigma_div > 1e3)
            nflagged_orig = len(np.where(weights[:, :, pol] == 0.0)[0])
            weights[:, flagged[0], pol] = 0.0
            nflagged_new = len(np.where(weights[:, :, pol] == 0.0)[0])
            median_val = np.nanmedian(amps[np.where(weights[:, :, pol] > 0.0)])
            if median_val < median_min or median_val > median_max:
                # Station has extreme median value
                logging.info('Flagged {0} (pol {1}) due to extreme median value '
                             '({2})'.format(ants[s], pol, median_val))
                weights[:, :, pol] = 0.0
            else:
                # Station is OK, flag bad points only
                prcnt = float(nflagged_new - nflagged_orig) / float(np.product(weights.shape[:-1])) * 100.0
                logging.info('Flagged {0:.1f}% of solutions for {1} (pol {2})'.format(prcnt, ants[s], pol))

    outQueue.put([s, weights])


def run( soltab, mode, maxFlaggedFraction=0.5, nSigma=5.0, maxStddev=None, ampRange=None, telescope='lofar', skipInternational=False, refAnt='', soltabExport='', ncpu=0 ):
    """
    This operation for LoSoTo implements a station-flagging procedure. Flags are time-independent.
    WEIGHT: compliant

    Parameters
    ----------
    mode: str
        Fitting algorithm: bandpass or resid. Bandpass mode clips amplitudes relative to a model bandpass (only LOFAR is currently supported). Resid mode clips residual phases or log(amplitudes).

    maxFlaggedFraction : float, optional
        This sets the maximum allowable fraction of flagged solutions above which the entire station is flagged.

    nSigma : float, optional
        This sets the number of standard deviations considered when outlier clipping is done.

    maxStddev : float, optional
        Maximum allowable standard deviation when outlier clipping is done. For phases, this should value
        should be in radians, for amplitudes in log(amp). If None (or negative), a value of 0.1 rad is
        used for phases and 0.01 for amplitudes.

    ampRange : array, optional
        2-element array of the median amplitude level to be acceptable, ampRange[0]: lower limit, ampRange[1]: upper limit.
        If None or [0, 0], a reasonable range for typical observations is used.

    telescope : str, optional
        Specifies the telescope if mode = 'bandpass'.

    skipInternational : str, optional
        If True, skip flagging of international LOFAR stations (only used if telescope = 'lofar')

    refAnt : str, optional
        If mode = resid, this sets the reference antenna for phase solutions, by default None.

    soltabExport : str, optional
        Soltab to export station flags to. Note: exported flags are not time- or frequency-dependent.

    ncpu : int, optional
        Number of cpu to use, by default all available.
    """

    logging.info("Flagging on soltab: "+soltab.name)

    # input check
    if refAnt == '':
        refAnt = None
    if soltabExport == '':
        soltabExport = None
    if mode is None or mode.lower() not in ['bandpass', 'resid']:
        logging.error('Mode must be one of bandpass or resid')
        return 1
    solType = soltab.getType()
    if maxStddev is None or maxStddev <= 0.0:
        if solType == 'phase':
            maxStddev = 0.1  # in radians
        else:
            maxStddev = 0.01  # in log10(amp)

    # Axis order must be [time, ant, freq, pol], so reorder if necessary
    axis_names = soltab.getAxesNames()
    if ('freq' not in axis_names or
        'time' not in axis_names or 'ant' not in axis_names):
        logging.error("Currently, flagstation requires the following axes: "
                      "freq, time, and ant.")
        return 1
    freq_ind = axis_names.index('freq')
    time_ind = axis_names.index('time')
    ant_ind = axis_names.index('ant')
    if 'pol' in axis_names:
        has_pol_axis = True
        pol_ind = axis_names.index('pol')
        if 'dir' in axis_names:
            dir_ind = axis_names.index('dir')
            vals_arraytmp = soltab.val[:].transpose([time_ind, ant_ind, freq_ind, pol_ind, dir_ind])
            weights_arraytmp = soltab.weight[:].transpose([time_ind, ant_ind, freq_ind, pol_ind, dir_ind])
        else:
            vals_arraytmp = soltab.val[:].transpose([time_ind, ant_ind, freq_ind, pol_ind])
            weights_arraytmp = soltab.weight[:].transpose([time_ind, ant_ind, freq_ind, pol_ind])
    else:
        has_pol_axis = False
        if 'dir' in axis_names:
            dir_ind = axis_names.index('dir')
            vals_arraytmp = soltab.val[:].transpose([time_ind, ant_ind, freq_ind, dir_ind])
            weights_arraytmp = soltab.weight[:].transpose([time_ind, ant_ind, freq_ind, dir_ind])
            vals_arraytmp = np.expand_dims(vals_arraytmp, axis=3)
            weights_arraytmp = np.expand_dims(weights_arraytmp, axis=3)
        else:
            vals_arraytmp = soltab.val[:].transpose([time_ind, ant_ind, freq_ind])
            weights_arraytmp = soltab.weight[:].transpose([time_ind, ant_ind, freq_ind])
            vals_arraytmp = np.expand_dims(vals_arraytmp, axis=-1)
            weights_arraytmp = np.expand_dims(weights_arraytmp, axis=-1)

    # Check for NaN solutions and flag
    flagged = np.where(np.isnan(vals_arraytmp))
    weights_arraytmp[flagged] = 0.0

    if mode == 'bandpass':
        if solType != 'amplitude':
            logging.error("Soltab must be of type amplitude for bandpass mode.")
            return 1

        # Fill the queue
        if 'dir' in axis_names:
            for d, dirname in enumerate(soltab.dir):
                mpm = multiprocManager(ncpu, _flag_bandpass)
                for s in range(len(soltab.ant)):
                    if ('CS' not in soltab.ant[s] and 'RS' not in soltab.ant[s] and
                        skipInternational and telescope.lower() == 'lofar'):
                        continue
                    mpm.put([soltab.freq[:], vals_arraytmp[:, s, :, :, d], weights_arraytmp[:, s, :, :, d],
                             telescope, nSigma, ampRange, maxFlaggedFraction, maxStddev, False, soltab.ant[:], s])
                mpm.wait()
                for (s, w) in mpm.get():
                    weights_arraytmp[:, s, :, :, d] = w
        else:
            mpm = multiprocManager(ncpu, _flag_bandpass)
            for s in range(len(soltab.ant)):
                if ('CS' not in soltab.ant[s] and 'RS' not in soltab.ant[s] and
                    skipInternational and telescope.lower() == 'lofar'):
                    continue
                mpm.put([soltab.freq[:], vals_arraytmp[:, s, :, :], weights_arraytmp[:, s, :, :],
                         telescope, nSigma, ampRange, maxFlaggedFraction, maxStddev, False, soltab.ant[:], s])
            mpm.wait()
            for (s, w) in mpm.get():
                weights_arraytmp[:, s, :, :] = w

        # Make sure that fully flagged stations have all pols flagged
        for s in range(len(soltab.ant)):
            for p in range(len(soltab.pol)):
                if np.all(weights_arraytmp[:, s, :, p] == 0.0):
                    weights_arraytmp[:, s, :, :] = 0.0
                    break

        # Write new weights
        if 'dir' in axis_names:
            weights_array = weights_arraytmp.transpose([time_ind, ant_ind, freq_ind, pol_ind, dir_ind])
        else:
            weights_array = weights_arraytmp.transpose([time_ind, ant_ind, freq_ind, pol_ind])
        soltab.setValues(weights_array, weight=True)
        soltab.addHistory('FLAGSTATION (mode=bandpass, telescope={0}, maxFlaggedFraction={1}, '
                          'nSigma={2})'.format(telescope, maxFlaggedFraction, nSigma))
    else:
        if solType not in ['phase', 'amplitude']:
            logging.error("Soltab must be of type phase or amplitude for resid mode.")
            return 1

        # Subtract reference phases
        if refAnt is not None:
            if solType != 'phase':
                logging.error('Reference possible only for phase solution tables. Ignoring referencing.')
            else:
                if refAnt == 'nearest':
                    for i, antToRef in enumerate(soltab.getAxisValues('ant')):
                        # get the closest antenna
                        antDists = soltab.getSolset().getAntDist(antToRef) # this is a dict
                        for badAnt in soltab._getFullyFlaggedAnts(): del antDists[badAnt] # remove bad ants
                        reference = list(antDists.keys())[list(antDists.values()).index( sorted(antDists.values())[1] ) ] # get the second closest antenna (the first is itself)
                        refInd = soltab.getAxisValues('ant', ignoreSelection=True).tolist().index(reference)
                        if 'dir' in axis_names:
                            vals_arrayref = vals_arraytmp[:, refInd, :, :, :].copy()
                        else:
                            vals_arrayref = vals_arraytmp[:, refInd, :, :].copy()
                        if 'dir' in axis_names:
                            vals_arraytmp[:, i, :, :, :] -= vals_arrayref
                        else:
                            vals_arraytmp[:, i, :, :] -= vals_arrayref
                else:
                    ants = soltab.getAxisValues('ant')
                    if refAnt not in ants:
                        logging.warning('Reference antenna '+refAnt+' not found. Using: '+ants[0])
                        refAnt = ants[0]
                    refInd = ants.tolist().index(refAnt)
                    if 'dir' in axis_names:
                        vals_arrayref = vals_arraytmp[:, refInd, :, :, :].copy()
                    else:
                        vals_arrayref = vals_arraytmp[:, refInd, :, :].copy()
                    for i in range(len(soltab.ant)):
                        if 'dir' in axis_names:
                            vals_arraytmp[:, i, :, :, :] -= vals_arrayref
                        else:
                            vals_arraytmp[:, i, :, :] -= vals_arrayref

        # Fill the queue
        if 'dir' in axis_names:
            for d, dirname in enumerate(soltab.dir):
                mpm = multiprocManager(ncpu, _flag_resid)
                for s in range(len(soltab.ant)):
                    mpm.put([vals_arraytmp[:, s, :, :, d], weights_arraytmp[:, s, :, :, d],
                             solType, nSigma, maxFlaggedFraction, maxStddev, soltab.ant[:], s])
                mpm.wait()
                for (s, w) in mpm.get():
                    weights_arraytmp[:, s, :, :, d] = w
        else:
            mpm = multiprocManager(ncpu, _flag_resid)
            for s in range(len(soltab.ant)):
                mpm.put([vals_arraytmp[:, s, :, :], weights_arraytmp[:, s, :, :], solType,
                         nSigma, maxFlaggedFraction, maxStddev, soltab.ant[:], s])
            mpm.wait()
            for (s, w) in mpm.get():
                weights_arraytmp[:, s, :, :] = w

        # Make sure that fully flagged stations have all pols flagged
        if has_pol_axis:
            npol = len(soltab.pol)
        else:
            npol = 1
        for s in range(len(soltab.ant)):
            for p in range(npol):
                if np.all(weights_arraytmp[:, s, :, p] == 0.0):
                    weights_arraytmp[:, s, :, :] = 0.0
                    break

        # Write new weights
        if 'dir' in axis_names:
            if has_pol_axis:
                weights_array = weights_arraytmp.transpose([time_ind, ant_ind, freq_ind, pol_ind, dir_ind])
            else:
                weights_array = weights_arraytmp[:, :, :, 0, :].transpose([time_ind, ant_ind, freq_ind, dir_ind])
        else:
            if has_pol_axis:
                weights_array = weights_arraytmp.transpose([time_ind, ant_ind, freq_ind, pol_ind])
            else:
                weights_array = weights_arraytmp[:, :, :, 0].transpose([time_ind, ant_ind, freq_ind])
        soltab.setValues(weights_array, weight=True)
        soltab.addHistory('FLAGSTATION (mode=resid, maxFlaggedFraction={0}, '
                          'nSigma={1})'.format(maxFlaggedFraction, nSigma))

    if soltabExport is not None:
        # Transfer station flags to soltabExport
        solset = soltab.getSolset()
        soltabexp = solset.getSoltab(soltabExport)
        axis_namesexp = soltabexp.getAxesNames()

        for stat in soltabexp.ant:
            if stat in soltab.ant:
                s = soltab.ant[:].tolist().index(stat)
                if 'pol' in axis_namesexp:
                    for pol in soltabexp.pol:
                        if pol in soltab.pol:
                            soltabexp.setSelection(ant=stat, pol=pol)
                            p = soltab.pol[:].tolist().index(pol)
                            if np.all(weights_arraytmp[:, s, :, p] == 0):
                                soltabexp.setValues(np.zeros(soltabexp.weight.shape), weight=True)
                else:
                    soltabexp.setSelection(ant=stat)
                    if np.all(weights_arraytmp[:, s, :, :] == 0):
                        soltabexp.setValues(np.zeros(soltabexp.weight.shape), weight=True)
        soltabexp.addHistory('WEIGHT imported by FLAGSTATION from '+soltab.name+'.')

    return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This tool is used to convert a parmdb into a H5parm format.
It can be run on a globaldb created with parmdb_collector.py
or on a single SB which contains the necessary
sky/instrument/ANTENNA/FIELD tables.
It can also load a dictionary from a json or pickle file with the keys 
instrumentdb (array with file names), antenna, field and skydb containing the 
corresponding file names.
It handles Gain/DirectionalGain/RotationAngle/
           Clock/TEC/CommonRotationAngle/CommonScalarPhase/CommonScalarAmpitude solution types.
"""
# Authors:
# Francesco de Gasperin
_author = "Francesco de Gasperin (astro@voo.it)"

import sys, os, glob
import socket
from losoto import _version
from losoto import _logging
from losoto._importer import create_h5parm

    
if __name__=='__main__':
    # Options
    import optparse
    opt = optparse.OptionParser(usage='%prog [-v] <H5parm> <globaldb/SBname> \n'\
                    +_author, version='%prog '+_version.__version__)
    opt.add_option('-V', '--verbose', help='Go Vebose! (default=False)', action='store_true', default=False)
    opt.add_option('-s', '--solset', help='Solution-set name (default=sol###)', type='string', default=None)
    opt.add_option('-i', '--instrument', help='Name of the instrument table (default=instrument*)', type='string', default='instrument*')
    opt.add_option('-c', '--complevel', help='Compression level from 0 (no compression, fast) to 9 (max compression, slow) (default=5)', type='int', default='5')
    (options, args) = opt.parse_args()

    # Check options
    if len(args) != 2:
        opt.print_help()
        sys.exit()

    # log
    logger = _logging.Logger('info')
    logging = _logging.logger
    if options.verbose: logger.set_level("debug")

    h5parmFile = args[0]
    logging.info("H5parm filename = "+h5parmFile)
    
    # Common options
    complevel = options.complevel
    solsetName = options.solset

    input_file = args[1]
    
    if input_file.endswith(".json"):
        try:
            import json
            logging.info("Loading json file: {}".format(input_file))
            files = json.load(open(input_file,"r"))
            instrumentdbFiles = [str(f) for f in files["instrumentdb"]]
            antennaFile = str(files["antenna"])
            fieldFile = str(files["field"])
            skydbFile = str(files["skydb"])
            globaldbFile = None
        except:
            logging.critical('Loading failed')
    elif input_file.endswith(".pckl"):
        try:
            import pickle
            logging.info("Loading pickle file: {}".format(input_file))
            files = pickle.load(open(input_file,"r"))
            instrumentdbFiles = files["instrumentdb"]
            antennaFile = files["antenna"]
            fieldFile = files["field"]
            skydbFile = files["skydb"]
            globaldbFile = None
        except:
            logging.critical('Loading failed')
    else:
        globaldbFile = input_file
        if not os.path.exists(globaldbFile):
            logging.critical('Input globaldb/SB file not found.')
            sys.exit(1)
        logging.info("globaldb filename = "+globaldbFile)
        
        # Load the path of the files
        antennaFile = os.path.join(globaldbFile,'ANTENNA')
        fieldFile = os.path.join(globaldbFile,'FIELD')
        skydbFile = os.path.join(globaldbFile,'sky')

        # Make a list of all available instrument tables (only 1 for a standard MS)
        instrumentdbFiles = [ instrumentdbFile for instrumentdbFile in \
            glob.glob(os.path.join(globaldbFile,options.instrument)) \
            if os.path.isdir(instrumentdbFile) ]

    # Check antennaFile and fieldFile
    if not os.path.isdir(antennaFile):
        logging.critical('Missing ANTENNA table.')
        sys.exit(1)
    if not os.path.isdir(fieldFile):
        logging.critical('Missing FIELD table.')
        sys.exit(1)
    
    # Call the method that creates the h5parm file
    create_h5parm(instrumentdbFiles, antennaFile, fieldFile, skydbFile,
                  h5parmFile, complevel, solsetName, globaldbFile=globaldbFile,verbose=options.verbose)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading RESET module.')

def _run_parser(soltab, parser, step):
    dataVal = parser.getfloat( step, 'dataVal', -999. )

    parser.checkSpelling( step, soltab, ['dataVal'])
    return run(soltab, dataVal)

def run( soltab, dataVal=-999. ):
    """
    This operation reset all the selected solution values.
    WEIGHT: flag compliant, no need for weight

    Parameters
    ----------
    dataVal : float, optional
        If given set values to this number, otherwise uses 1 for amplitude and 0 for all other soltab types.
    """

    logging.info("Resetting soltab: "+soltab.name)

    solType = soltab.getType()

    if dataVal == -999.:
        if solType == 'amplitude':
            dataVal = 1.
        else:
            dataVal = 0.
    
    soltab.setValues(dataVal)

    soltab.addHistory('RESET')
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading DUPLICATE module.')

def _run_parser(soltab, parser, step):
    soltabOut = parser.getstr( step, 'soltabOut', '' )
    overwrite = parser.getbool( step, 'overwrite', False )

    parser.checkSpelling( step, soltab, ['soltabOut','overwrite'])
    return run(soltab, soltabOut, overwrite)

def run( soltab, soltabOut='', overwrite=False):
    """
    Duplicate a table

    Parameters
    ----------
    soltabOut : str, optional
        Output table name. By default choose next available from table type.
    overwrite : bool, optional
        Overwrite soltabOut if it already exists?
    """

    if soltabOut == '':
        soltabOut = None

    solset = soltab.getSolset()
    if soltabOut in solset.getSoltabNames() and overwrite:
        logging.info('Overwriting soltabOut {}'.format(soltabOut))
        solset.getSoltab(soltabOut).delete()

    soltabout = solset.makeSoltab(soltype = soltab.getType(), soltabName = soltabOut, axesNames=soltab.getAxesNames(),
        axesVals=[soltab.getAxisValues(axisName) for axisName in soltab.getAxesNames()],
        vals=soltab.getValues(retAxesVals = False), weights=soltab.getValues(weight = True, retAxesVals = False))
    # parmdbType=soltab.obj._v_attrs['parmdb_type'] # deprecated

    logging.info('Duplicate %s -> %s' % (soltab.name, soltabout.name) )

    soltabout.addHistory('DUPLICATE from table %s' % (soltab.name))
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading REWEIGHT module.')

def _run_parser(soltab, parser, step):
    mode = parser.getstr( step, 'mode', 'uniform' )
    weightVal = parser.getfloat( step, 'weightVal', 1. )
    nmedian = parser.getint( step, 'nmedian', 3 )
    nstddev = parser.getint( step, 'nstddev', 251 )
    soltabImport = parser.getstr( step, 'soltabImport', '' )
    flagBad = parser.getbool( step, 'flagBad', False )
    ncpu = parser.getint( '_global', 'ncpu', 0 )

    parser.checkSpelling( step, soltab, ['mode', 'weightVal', 'nmedian', 'nstddev', 'soltabImport', 'flagBad'])
    return run(soltab, mode, weightVal, nmedian, nstddev, soltabImport, flagBad, ncpu)


def _rolling_window_lastaxis(a, window):
    """Directly taken from Erik Rigtorp's post to numpy-discussion.
    <http://www.mail-archive.com/numpy-discussion@scipy.org/msg29450.html>"""
    import numpy as np

    if window < 1:
        raise ValueError("`window` must be at least 1.")
    if window > a.shape[-1]:
        raise ValueError("`window` is too long.")
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def _nancircstd(samples, axis=None, is_phase=True):
    """
    Compute the circular standard deviation

    Based on scipy.stats.circstd

    Parameters
    ----------
    imag : array_like
        Input array.
    axis : int, optional
        Axis along which standard deviations are computed.  The default is
        to compute the standard deviation of the flattened array.
    is_phase : bool, optional
        If True, samples are assumed to be phases. If False, they are assumed
        to be either real or imaginary values

    Returns
    -------
    circstd : float
        Circular standard deviation.
    """
    import numpy as np

    if is_phase:
        x1 = np.sin(samples)
        x2 = np.cos(samples)
    else:
        x1 = samples
        x2 = np.sqrt(1.0 - x1**2)
    R = np.hypot(np.nanmean(x1, axis=axis), np.nanmean(x2, axis=axis))

    return np.sqrt(-2*np.log(R))


def _estimate_weights_window(sindx, vals, nmedian, nstddev, stype, outQueue):
    """
    Set weights using a median-filter method

    Parameters
    ----------
    sindx: int
        Index of station
    vals: array
        Array of values
    nmedian: odd int
        Size of median time window
    nstddev: odd int
        Size of stddev time window
    stype: str
        Type of values (e.g., 'phase')

    """
    import numpy as np
    from scipy.ndimage import generic_filter

    pad_width = [(0, 0)] * len(vals.shape)
    pad_width[-1] = (int((nmedian-1)/2), int((nmedian-1)/2))
    if stype == 'phase' or stype == 'rotation':
        # Median smooth and subtract to de-trend
        if nmedian > 0:
            # Convert to real/imag
            real = np.cos(vals)
            pad_real = np.pad(real, pad_width, 'constant', constant_values=(np.nan,))
            med_real = np.nanmedian(_rolling_window_lastaxis(pad_real, nmedian), axis=-1)
            real -= med_real
            real[real < -1.0] = -1.0
            real[real > 1.0] = 1.0

            imag = np.sin(vals)
            pad_imag = np.pad(imag, pad_width, 'constant', constant_values=(np.nan,))
            med_imag = np.nanmedian(_rolling_window_lastaxis(pad_imag, nmedian), axis=-1)
            imag -= med_imag
            imag[imag < -1.0] = -1.0
            imag[imag > 1.0] = 1.0

            # Calculate standard deviations
            pad_width[-1] = (int((nstddev-1)/2), int((nstddev-1)/2))
            pad_real = np.pad(real, pad_width, 'constant', constant_values=(np.nan,))
            stddev1 = _nancircstd(_rolling_window_lastaxis(pad_real, nstddev), axis=-1, is_phase=False)
            pad_imag = np.pad(imag, pad_width, 'constant', constant_values=(np.nan,))
            stddev2 = _nancircstd(_rolling_window_lastaxis(pad_imag, nstddev), axis=-1, is_phase=False)
            stddev = stddev1 + stddev2
        else:
            phase = normalize_phase(vals)

            # Calculate standard deviation
            pad_width[-1] = (int((nstddev-1)/2), int((nstddev-1)/2))
            pad_phase = np.pad(phase, pad_width, 'constant', constant_values=(np.nan,))
            stddev = _nancircstd(_rolling_window_lastaxis(pad_phase, nstddev), axis=-1)
    else:
        if stype == 'amplitude':
            # Assume lognormal distribution for amplitudes
            vals = np.log(vals)

        # Median smooth and subtract to de-trend
        if nmedian > 0:
            pad_vals = np.pad(vals, pad_width, 'constant', constant_values=(np.nan,))
            med = np.nanmedian(_rolling_window_lastaxis(pad_vals, nmedian), axis=-1)
            vals -= med

        # Calculate standard deviation in larger window
        pad_width[-1] = (int((nstddev-1)/2), int((nstddev-1)/2))
        pad_vals = np.pad(vals, pad_width, 'constant', constant_values=(np.nan,))
        stddev = np.nanstd(_rolling_window_lastaxis(pad_vals, nstddev), axis=-1)

    # Check for periods where standard deviation is zero or NaN and replace
    # with min value to prevent inf in the weights. Also limit weights to
    # float16
    zero_scatter_ind = np.where(np.logical_or(np.isnan(stddev), stddev == 0.0))
    if len(zero_scatter_ind[0]) > 0:
        good_ind = np.where(~np.logical_or(np.isnan(stddev), stddev == 0.0))
        stddev[zero_scatter_ind] = np.min(stddev[good_ind])
    if nmedian > 0:
        fudge_factor = 2.0  # factor to compensate for smoothing
    else:
        fudge_factor = 1.0
    w = 1.0 / np.square(stddev*fudge_factor)

    # Rescale to fit in float16
    float16max = 65504.0
    if np.max(w) > float16max:
        w *= float16max / np.max(w)

    outQueue.put([sindx, w])


def run( soltab, mode='uniform', weightVal=1., nmedian=3, nstddev=251,
    soltabImport='', flagBad=False, ncpu=0 ):
    """
    Change the the weight values.

    Parameters
    ----------
    mode : str, optional
        One of 'uniform' (single value), 'window' (sliding window in time), or 'copy' (copy from another table), by default 'uniform'.
    weightVal : float, optional
        Set weights to this values (0=flagged), by default 1.
    nmedian : odd int, optional
        Median window size in number of timeslots for 'window' mode.
        If nonzero, a median-smoothed version of the input values is
        subtracted to detrend them. If 0, no smoothing or subtraction is
        done, by default 3.
    nstddev : odd int, optional
        Standard deviation window size in number of timeslots for 'window' mode, by default 251.
    soltabImport : str, optional
        Name of a soltab. Copy weights from this soltab (must have same axes shape), by default none.
    flagBad : bool, optional
        Re-apply flags to bad values (1 for amp, 0 for other tables), by default False.
    """

    import numpy as np

    logging.info("Reweighting soltab: "+soltab.name)

    if mode == 'copy':
        if soltabImport == '':
            logging.error('In copy mode a soltabImport must be specified.')
            return 1
        solset = soltab.getSolset()
        soltabI = solset.getSoltab(soltabImport)
        soltabI.selection = soltab.selection

        weights, axes = soltab.getValues(weight = True)
        weightsI, axesI = soltabI.getValues(weight = True)
        if list(axes.keys()) != list(axesI.keys()) or weights.shape != weightsI.shape:
            logging.error('Impossible to merge: two tables have with different axes values.')
            return 1
        weightsI[ np.where(weights == 0) ] = 0.
        soltab.setValues(weightsI, weight=True)
        soltab.addHistory('WEIGHT imported from '+soltabI.name+'.')

    elif mode == 'uniform':
        soltab.addHistory('REWEIGHTED to '+str(weightVal)+'.')
        soltab.setValues(weightVal, weight=True)

    elif mode == 'window':
        if nmedian !=0 and nmedian % 2 == 0:
            logging.error('nmedian must be odd')
            return 1
        if nstddev % 2 == 0:
            logging.error('nstddev must be odd')
            return 1

        tindx = soltab.axesNames.index('time')
        antindx = soltab.axesNames.index('ant')
        vals = soltab.val[:].swapaxes(antindx, 0)
        if tindx == 0:
            tindx = antindx
        mpm = multiprocManager(ncpu, _estimate_weights_window)
        for sindx, sval in enumerate(vals):
            if np.all(sval == 0.0) or np.all(np.isnan(sval)):
                # skip reference station
                continue
            mpm.put([sindx, sval.swapaxes(tindx-1, -1), nmedian, nstddev, soltab.getType()])
        mpm.wait()
        weights = np.ones(vals.shape)
        for (sindx, w) in mpm.get():
            weights[sindx, :] = w.swapaxes(-1, tindx-1)
        weights = weights.swapaxes(0, antindx)

        soltab.addHistory('REWEIGHTED using sliding window with nmedian={0} '
            'and nstddev={1} timeslots'.format(nmedian, nstddev))
        soltab.setValues(weights, weight=True)

    if flagBad:
        weights = soltab.getValues(weight = True, retAxesVals = False)
        vals = soltab.getValues(retAxesVals = False)
        if soltab.getType() == 'amplitude': weights[np.where(vals == 1)] = 0
        else: weights[np.where(vals == 0)] = 0
        soltab.setValues(weights, weight=True)

    return 0



#!/usr/bin/python
# -*- coding: utf-8 -*-
import numpy as np
import numpy.ma as ma
import sys
from multiprocessing import Pool
from losoto._logging import logger as logging

has_fitting=True
try:
    import casacore.tables # must be loaded before expion - used in other operations as lofarbeam
    from lofar.expion import baselinefitting as fitting
except:
#    logging.error('No lofar.expion present, clock/tec separation not active.')
    logging.debug('No lofar.expion present, clock/tec separation maybe not optimal')
    has_fitting=False
light_speed=299792458.
# from pylab import *

def ClockTEC_3rdorder_func(xarray, par):
    """clock tec fitting withextra parameter for third order ionospheric effects at lowest frequencies"""
    delay = np.array([par[1] * 1e-9]).flatten()  # in ns, array has dimension 1, even if scalar
    delayfact = 2 * np.pi * delay[:, np.newaxis] * xarray
    TEC = np.array([par[0]]).flatten()  # dTEC in TECU
    drefract = -8.4479745e9 * TEC[:, np.newaxis] / xarray
    TEC3rd = np.array([par[2]]).flatten()  # 3rd_order
    d3rd = light_speed**3* TEC3rd[:, np.newaxis] / xarray**3

    return drefract[:, np.newaxis,np.newaxis, :] + delayfact[np.newaxis,:,np.newaxis, :]+d3rd[np.newaxis,np.newaxis,:, :]

def ClockTECfunc(xarray, par):
    delay = np.array([par[1] * 1e-9]).flatten()  # in ns, array has dimension 1, even if scalar
    delayfact = 2 * np.pi * delay[:, np.newaxis] * xarray
    TEC = np.array([par[0]]).flatten()  # dTEC in TECU
    drefract = -8.4479745e9 * TEC[:, np.newaxis] / xarray
    if len(par) > 2:
        return drefract[:, np.newaxis, :] + delayfact[np.newaxis] + par[2]  # returns nTEC x nClock x nFreq

    return drefract[:, np.newaxis, :] + delayfact


def ClockTECfuncAllStations(xarray, par):
    delay = np.array([par[1] * 1e-9]).flatten()  # in ns, array has dimension 1, even if scalar
    delayfact = 2 * np.pi * delay[:, np.newaxis] * xarray
    # print "delayfact",delayfact.shape
    TEC = np.array([par[0]]).flatten()  # dTEC in TECU
    drefract = -8.4479745e9 * TEC[:, np.newaxis] / xarray
    # print "refract",drefract.shape
    if len(par) > 2:
        return drefract + delayfact + par[2][:, np.newaxis]  # returns nTEC x nClock x nFreq

    return drefract + delayfact


def getInitClock(data, freq):
    nF = freq.shape[0]
    avgdata = np.ma.sum(np.cos(data) + 1.j * np.sin(data), axis=0).swapaxes(0, -2)
    avgdata = np.ma.arctan2(np.imag(avgdata), np.real(avgdata))
    nSt = avgdata.shape[0]
    npol = avgdata.shape[2]
    for ist in range(nSt):
        for pol in range(npol):
            mymask = avgdata[ist, :, pol].mask
            if not hasattr(mymask, '__len__'):
                mymask = np.ones(avgdata[ist, :, pol].shape, dtype=bool) * mymask
            if avgdata[ist, :, pol].count() < 1:
                avgdata[ist, :, pol].mask[0] = False
            # logging.debug("mask station %d pol %d "%(ist,pol) +str(mymask))
            # logging.debug("average data station %d pol %d "%(ist,pol) +str(avgdata[ist,:,pol]))
            avgdata[ist, :, pol][~mymask] = np.float32(unwrapSparsePhases(avgdata[ist, :, pol][~mymask],freq[~mymask]))
            # logging.debug("average unwrapped data station %d pol %d "%(ist,pol) +str(avgdata[ist,:,pol]))
            # logging.debug("remainder " +str(np.remainder(avgdata[ist,:,pol]+np.pi,2*np.pi)-np.pi))
    A = np.ones((nF, 2), dtype=np.float)
    A[:, 1] = freq * 2 * np.pi * 1e-9
    return np.ma.dot(np.linalg.inv(np.dot(A.T, A)), np.ma.dot(A.T, avgdata).swapaxes(0, -2))

def unwrapSparsePhases(phases,freqs,doFlag=False):
    '''unwrap phases, using frequency coverage'''
    testwraps=np.arange(-25,26,.1)
    dclock=testwraps*1e9/(freqs[-1]-freqs[0])
    A = np.ones((freqs.shape[0], 2), dtype=np.float)
    A[:, 1] = freqs * 2 * np.pi * 1e-9
    testdata=np.zeros_like(dclock)
    testdata=np.array([testdata,dclock])
    fitdata=np.ma.dot(testdata.T,A.T)
    offsets=np.ma.average(np.ma.remainder(phases[np.newaxis]-fitdata+0.5*np.pi,np.pi)-0.5*np.pi,axis=-1)
    fitdata=fitdata-offsets[:,np.newaxis]
    wraps=np.ma.round((phases[np.newaxis]-fitdata)/(2*np.pi))
    nphases=phases[np.newaxis]-wraps*2*np.pi
    myvar1=np.ma.var((nphases-fitdata[np.newaxis]),axis=-1)
    idx=np.argmin(myvar1)

    dclock=dclock[idx]
    fitdata=np.ma.dot(np.array([0,dclock]),A.T)+offsets[idx]
    return unwrapPhases(phases,freqs,fitdata=fitdata,doFlag=doFlag)


def unwrapPhases(phases,freqs,fitdata=None,maskrange=10,doFlag=True,flagfitdata=False):
    '''unwrap phases, remove jumps and get best match with fitdata'''
    mymask=phases.mask
    freqsstep=np.min(freqs[1:]-freqs[:-1])
    nfreq=np.arange(freqs[0],freqs[-1]+0.5*freqsstep,freqsstep)
    freqidx=np.argmin(np.abs(freqs[np.newaxis]-nfreq[:,np.newaxis]),axis=0)
    maxgap=int(np.round(3.e6/freqsstep))
    maskrange=max(maskrange,maxgap)
    unmasked=np.zeros_like(nfreq)
    full_mask=np.ones(unmasked.shape,dtype=bool)
    full_mask[freqidx]=mymask
    for nriter in range(2):
        if fitdata is not None and fitdata.shape == phases.shape:
            wraps=np.ma.round((phases-fitdata)/(2*np.pi))
            phases-=wraps*2*np.pi
            unmasked[freqidx]=np.copy(np.array(phases))
            unmasked[np.isnan(unmasked)]=0.

        if fitdata is None:
            unmasked[freqidx]=np.copy(np.array(phases))
            unmasked[np.isnan(unmasked)]=0.
            unmasked=np.unwrap(unmasked)
            wraps=np.ma.round((phases-unmasked[freqidx])/(2*np.pi))
            phases-=wraps*2*np.pi
        maskpoints=np.where(full_mask)[0]
        if maskpoints.shape[0]>0:
            Atmp=np.ones((maskrange,2),dtype=np.float64)
            Atmp[:,1]=np.arange(maskrange)
            Atmpinv=np.linalg.inv(np.dot(Atmp.T,Atmp))
        doreverse=False
        for i in maskpoints:
            if i<maskrange:
                doreverse=True
            if i>=maskrange:
                #print "changing unmasked[",i,"]:",unmasked[i],"into",
                unmasked=np.unwrap(unmasked)
                #print unmasked[i]
                unmasked[i]=np.dot(np.dot(Atmpinv,np.dot(Atmp.T,unmasked[i-maskrange:i])),[1,maskrange])
                #print unmasked[i]
        if doreverse:
            for i in maskpoints[::-1]:
                if i<unmasked.shape[0]-1-maskrange:
                    unmasked[i:]=np.unwrap(unmasked[i:])
                    unmasked[i]=np.dot(np.dot(Atmpinv,np.dot(Atmp.T,unmasked[i+1:i+maskrange+1])),[1,-1])
        unmasked=np.unwrap(unmasked)
        if doFlag:
            # detect jumps and remove them
            diffdata=unmasked[1:]-unmasked[:-1]
            #detect bad datapoints since they can destroy unwrapping (if the offset is close to np.pi)
            #wrapflags=np.logical_and(np.absolute(diffdata[:-1])>0.4*np.pi,np.absolute(diffdata[1:])>0.4*np.pi)
            wrapflags=np.logical_and(np.absolute(diffdata[:-1])>0.4*np.pi,np.absolute(diffdata[1:])>0.4*np.pi)
            #wrapflags=np.absolute(diffdata[:-1])>0.4*np.pi
            newmask=np.zeros_like(diffdata,dtype=bool)
            newmask[:-1]=wrapflags
            diffdata=np.ma.array(diffdata,mask=newmask)
            # use 2.5 pi for calculating jumps, tomake sureyou have a real 2pi jump,instead of a sequence of 2 bad datapoints with order 1pi jump. yes I have seen those in LBA calibrator data
            unmasked[1:]-=np.ma.cumsum(np.ma.round(diffdata/(2.5*np.pi)))*2*np.pi
            #wrapflags=np.logical_and(np.absolute(np.ma.ediff1d(phases)[:-1])>0.2*np.pi,np.absolute(np.ma.ediff1d(phases)[1:])>0.2*np.pi)
            full_mask[1:-1]=np.logical_or(full_mask[1:-1],wrapflags)
            phases[:]=unmasked[freqidx]
            phases.mask=full_mask[freqidx]
            # get best match with fitdata
        if fitdata is None:
            #average around 0
            A=np.ma.zeros((freqs.shape[0],2),dtype=np.float64)
            A.mask=np.zeros((freqs.shape[0],2))
            A[:,1]=2*np.pi*1e-9*freqs
            A[:,0]=-8.44797245e9/freqs
            A.mask[:,0]=phases.mask[:]
            A.mask[:,1]=phases.mask[:]
            par=np.ma.dot(np.linalg.inv(np.ma.dot(A[:,:2].T,A[:,:2])),np.ma.dot(A[:,:2].T,phases))
            #print par
            tmpfitdata=np.ma.dot(par,A.T)
            wraps=np.ma.round((phases-tmpfitdata)/(2*np.pi))
            phases-=wraps*2*np.pi
            par=np.ma.dot(np.linalg.inv(np.ma.dot(A[:,:2].T,A[:,:2])),np.ma.dot(A[:,:2].T,phases))
            tmpfitdata=np.ma.dot(par,A.T)
            wraps=np.ma.round((phases-tmpfitdata)/(2*np.pi))
            phases-=wraps*2*np.pi
            wrapflags=np.absolute(phases-tmpfitdata)>0.4*np.pi
            full_mask[freqidx]=np.logical_or(full_mask[freqidx],wrapflags)
            phases.mask=full_mask[freqidx]
            #print "par agina",par,wraps,phases
            #phases-=np.ma.round(np.ma.average(phases)/(2*np.pi))*np.pi*2
        else:

            phases-=np.ma.round(np.ma.average(phases-fitdata)/(2*np.pi))*np.pi*2
            if flagfitdata:
                newmask=np.absolute(fitdata-phases)>0.4*np.pi
                phases.mask=np.logical_or(mymask,newmask)
        if not doFlag or np.sum(wrapflags)==0:
            return phases
    #TO DO, if residuals remain too large after unwrapping, do a brute force search
    return phases


def getInitPar(
    data,
    freqs,
    nrTEC=40,
    nrClock=40,
    nrthird=0,
    initsol=tuple()
    ):
    #decide if flagging shouldbe used when unwrapping, depends on frequency coverage
    avgfreqstep=np.average(freqs[1:]-freqs[:-1])
    if avgfreqstep>2.e6:
        #with this freqstep you can expect 1 pi jumps with a clock error of 200 ns
        doFlag=False
    else:
        doFlag=True
    if nrthird>0:
        A=np.ma.zeros((freqs.shape[0],3),dtype=np.float64)
        A[:,1]=2*np.pi*1e-9*freqs
        A[:,0]=-8.44797245e9/freqs
        A[:,2]=-1.e21/freqs**3
    else:
        A=np.ma.zeros((freqs.shape[0],2),dtype=np.float64)
        A[:,1]=2*np.pi*1e-9*freqs
        A[:,0]=-8.44797245e9/freqs
    a=np.mgrid[int(-nrTEC/2):int(nrTEC/2)+1,-int(nrClock/2):int(nrClock/2)+1]
    #a=np.mgrid[int(-nrTEC*50):int(nrTEC*2)+1,-int(nrClock*2):int(nrClock*2)+1]
    if len(initsol)>=2 and not (initsol[0]==0 and initsol[1]==0) and not (initsol[0]==-10 and initsol[1]==-10)  :
        #print "unwrapping with initsol",initsol
        fitdata=np.dot(initsol,A.T)
        data=unwrapPhases(data,freqs,fitdata,doFlag=doFlag)
    else:
        if doFlag:
#            logging.debug("unwrapping pahses %d"%(data.count()))
            data=unwrapPhases(data,freqs,doFlag=doFlag)
#            logging.debug("unwrapped pahses %d"%(data.count()))
        else:
            data=unwrapSparsePhases(data,freqs)
        steps = np.ma.dot(np.ma.dot(np.linalg.inv(np.ma.dot(A[:,:2].T, A[:,:2])), A[:,:2].T), 2 * np.pi * np.ones((freqs.shape[0], ), dtype=np.float))
        par=np.ma.dot(np.linalg.inv(np.ma.dot(A[:,:2].T,A[:,:2])),np.ma.dot(A[:,:2].T,data))
        #get parameters close to 0
        data-=np.round(np.average(np.round(par/steps)))*2*np.pi
        par=np.ma.dot(np.linalg.inv(np.ma.dot(A[:,:2].T,A[:,:2])),np.ma.dot(A[:,:2].T,data))
        nrTEC+=np.abs(np.round(par[0]/steps[0]))
        nrClock+=np.abs(np.round(par[1]/steps[1]))
    if data.mask.sum()<0.5*data.size:
        A=np.ma.array(A,mask=np.tile(data.mask,(A.shape[1],1)).T)
    steps = np.ma.dot(np.ma.dot(np.linalg.inv(np.ma.dot(A[:,:2].T, A[:,:2])), A[:,:2].T), 2 * np.pi * np.ones((freqs.shape[0], ), dtype=np.float))
    #get initial guess, first only for first two parameters
    par=np.ma.dot(np.linalg.inv(np.ma.dot(A[:,:2].T,A[:,:2])),np.ma.dot(A[:,:2].T,data))

    #print "intial guess",par,steps,"min 0",a[0][0,0]*steps[0]*0.01+par[0],"max 0",a[0][-1,0]*steps[0]*0.01+par[0],"min 1",a[1][0,0]*steps[1]*0.01+par[1],"max 1",a[1][0,-1]*steps[1]*0.01+par[1]
    bigdata=np.concatenate(tuple([a[i][np.newaxis,:]*steps[i]+par[i] for i  in range(2)]),axis=0).transpose(1,2,0)
    #bigdata=np.concatenate(tuple([a[i][np.newaxis,:]*steps[i]*0.25+par[i] for i  in range(2)]),axis=0).transpose(1,2,0)
    diffdata=np.dot(bigdata,A[:,:2].T)
    diffdata-=data[np.newaxis,np.newaxis]
    idx=np.unravel_index(np.argmin(np.ma.var(diffdata,axis=-1)),diffdata.shape[:-1])
    par=bigdata[idx]
    #print "final giess",par,bigdata,"diffffff",idx,np.ma.var(diffdata,axis=-1)
    fitdata=np.dot(par,A[:,:2].T)
    data=unwrapPhases(data,freqs,fitdata,doFlag=doFlag,flagfitdata=True)
    if data.mask.sum()<0.5*data.size:
        A=np.ma.array(A,mask=np.tile(data.mask,(A.shape[1],1)).T)
    #now add third parameter if needed:
    if nrthird>0:
        steps = np.ma.dot(np.ma.dot(np.linalg.inv(np.ma.dot(A.T, A)), A.T), 2 * np.pi * np.ones((freqs.shape[0], ), dtype=np.float))
        par=np.ma.dot(np.linalg.inv(np.ma.dot(A.T,A)),np.ma.dot(A.T,data))
        a=np.mgrid[max(-1,int(-nrTEC/2)):min(2,int(nrTEC/2)+1),max(-1,int(-nrClock/2)):min(2,int(nrClock/2)+1),-int(nrthird/2):int(nrthird/2)+1] #assume dTEC and dClock are already close
        bigdata=np.concatenate(tuple([a[i][np.newaxis,:]*steps[i]+par[i] for i  in range(3)]),axis=0).transpose(1,2,3,0)
        diffdata=np.ma.dot(bigdata,A.T)
        diffdata-=data[np.newaxis,np.newaxis,np.newaxis]
        idx=np.unravel_index(np.argmin(np.ma.var(diffdata,axis=-1)),diffdata.shape[:-1])
        par=bigdata[idx]
        fitdata=np.dot(par,A.T)
        data=unwrapPhases(data,freqs,fitdata,doFlag=doFlag)
    return par,data


def getClockTECFit(
    ph,
    freq,
    stations,
    initSol=[],
    returnResiduals=True,
    chi2cut=1e8,
    fit3rdorder=False,
    double_search_space=False
    ):
    nT = ph.shape[0]
    nF = freq.shape[0]
    nSt = ph.shape[2]
    data = ph
    tecarray = np.zeros((nT, nSt), dtype=np.float32)
    clockarray = np.zeros((nT, nSt), dtype=np.float32)

    if returnResiduals:
        residualarray = np.zeros((nT, nF, nSt), dtype=np.float32)
    if fit3rdorder:
        tec3rdarray= np.zeros((nT, nSt), dtype=np.float32)
    A = np.ones((nF, 2+fit3rdorder), dtype=np.float)
    A[:, 1] = freq * 2 * np.pi * 1e-9
    A[:, 0] = -8.44797245e9 / freq
    if fit3rdorder:
        A[:, 2] = -1e21 / freq**3
    steps = np.ma.dot(np.ma.dot(np.linalg.inv(np.ma.dot(A.T, A)), A.T), 2 * np.pi * np.ones((freq.shape[0], ), dtype=np.float))
    succes=False
    initprevsol=np.zeros(nSt,dtype=bool)
    nrFail=np.zeros(nSt,dtype=int)
    sol = np.zeros((nSt, 2+fit3rdorder), dtype=np.float)
    prevsol = np.zeros_like(sol)
    n3rd=0
    for itm in range(nT):
        datatmp=np.ma.copy(data[itm, :])
        if itm == 0 or not succes:
            for ist in range(nSt):
                if itm == 0 or not initprevsol[ist]:
                    if hasattr(initSol, '__len__') and len(initSol) > ist:
                        sol[ist,:initSol[ist].shape[0]]=initSol[ist]
                        ndt=1
                        ndtec=1
                        if fit3rdorder:
                            n3rd=1
                    else:
                        sol[ist]=0
                        if fit3rdorder:
                            n3rd=200
                        if 'CS' in stations[ist]:
                            ndt=4
                            ndtec=10
                            if 'LBA' in stations[ist]:
                                ndt=2
                                ndtec=40
                        else:
                            if 'RS' in stations[ist]:
                                ndt=200
                                ndtec=80
                            else:
                                # large TEC variation for EU stations
                                ndt=200
                                ndtec=160
                            if 'LBA' in stations[ist]:
                                 ndt=60
                                 # no init clock possible due to large TEC effect
                                 #stepsize of dtec is small
                                 ndtec=520
                else:
                    # further steps with non success
                    sol[ist, :] = prevsol[ist, :]
                    ndtec=min(nrFail[ist]+1,100)
                    if not 'CS' in stations[ist]:
                        ndt=min(nrFail[ist]+1,200)
                    else:
                        ndt=min(nrFail[ist]+1,4)
                    if fit3rdorder:
                        n3rd=min(nrFail[ist]+1,200)
                datatmpist = datatmp[:, ist]
                if datatmpist.count() / float(nF) > 0.5:
                    # do brutforce and update data, unwrp pdata,update flags
                    #if ist==23 or ist==25:
                    #logging.debug("Getting init par for time %d:station %d ntec %d ndt %d n3rd %d"%(itm,ist,ndtec,ndt,n3rd)+str(sol[ist]))
                    par,datatmp[:, ist] = getInitPar(datatmpist, freq,nrTEC=ndtec*(1+double_search_space),nrClock=ndt*(1+double_search_space),nrthird=n3rd*(1+double_search_space),initsol=sol[ist,:])
                    sol[ist, :] = par[:]
                #if itm%100==0:
                    #logging.debug("Getting init par for station %d:%d "%(itm,ist)+str(sol[ist]))
        for ist in range(nSt):
            #now do the real fitting
            datatmpist=datatmp[:,ist]
            if datatmpist.count() / float(nF) < 0.5:
                logging.debug("Too many data points flagged first t=%d st=%d flags=%d" % (itm,ist,datatmpist.count()) + str(sol[ist]))
                sol[ist] = [-10.,]*sol.shape[1]
                continue
            fitdata=np.dot(sol[ist],A.T)
            datatmpist=unwrapPhases(datatmpist,freq,fitdata)
            #if itm%100==0:
            #logging.debug(" init par for station itm %d:%d "%(itm,ist)+str(sol[ist]))
            mymask=datatmpist.mask
            maskedfreq=np.ma.array(freq,mask=mymask)
            A2=np.ma.array(A,mask=np.tile(mymask,(A.shape[1],1)).T)
            sol[ist] = np.ma.dot(np.linalg.inv(np.ma.dot(A2.T, A2)), np.ma.dot(A2.T, datatmpist)).T
            if initprevsol[ist] and np.abs((sol[ist,1]-prevsol[ist,1])/steps[1])>0.5 and (np.abs((sol[ist,1]-prevsol[ist,1])/steps[1])>0.75 or np.abs(np.sum((sol[ist]-prevsol[ist])/steps,axis=-1))>0.5*A2.shape[0]):
                #logging.debug("removing jumps, par for station %d , itm %d"%(ist,itm)+str(sol[ist])+str(prevsol[ist])+str(steps))
                sol[ist,:]-=np.round((sol[ist,1]-prevsol[ist,1])/steps[1])*steps
                #logging.debug("removed jumps, par for station %d "%ist+str(sol[ist])+str(prevsol[ist]))
            #if itm%100==0:
            #logging.debug("par for station itm %d:%d "%(itm,ist)+str(sol[ist]))
         # calculate chi2 per station
        residual = data[itm] - np.dot(A, sol.T)
        tmpresid = residual - residual[:, 0][:, np.newaxis]  # residuals relative to station 0
        residual = np.ma.remainder(tmpresid + np.pi, 2 * np.pi) - np.pi
        chi2 = np.ma.sum(np.square(np.degrees(residual)), axis=0) / nF

        if returnResiduals:
            residualarray[itm] = residual

        chi2select = np.logical_or(np.array(chi2 > chi2cut), sol[:, 0] < -5) # select bad points
        chi2select = np.logical_or(chi2select,initprevsol*np.sum(np.abs((sol-prevsol)/steps),axis=-1)>(0.3*sol.shape[1]*(1+nrFail))) #also discard data where there is a "half" jump for any parameter wrst the previous solution (if previous solution exists). Multiply with number of fails, since after a large number of fails there is no clear match with the previous solution expected anyway...
        if np.any(chi2select):
            #logging.debug('high chi2 of fit, itm: %d %d ' % (itm, np.sum(chi2select)) + str(sol[chi2select]) + 'prevsol'+str(prevsol[chi2select])+'stations:' + str(np.arange(nSt)[chi2select]) + ' chi2 ' + str(chi2[chi2select]))
            logging.debug('High chi2 of fit, itm: %d %d ' % (itm, np.sum(chi2select)) + 'stations:' + str(np.arange(nSt)[chi2select]))
            succes = False
            nrFail[chi2select] += 1
            nrFail[~chi2select] = 0
            prevsol[np.logical_and(initprevsol == False,chi2select==False)] = sol[np.logical_and(initprevsol == False,chi2select==False)]  # compensate missing prevsol at first rounds
            prevsol[~chi2select] = 0.5 * prevsol[~chi2select] + 0.5 * sol[~chi2select]  # init solution to 0.5 * this solution + 0.5 previous solution
            initprevsol[~chi2select] = True # once is True it never becomes False
        else:
            # prevsol=np.copy(sol)
            prevsol[initprevsol==False] = sol[initprevsol == False]  # compensate missing prevsol at first rounds
            prevsol = 0.5 * prevsol + 0.5 * np.copy(sol)
            succes = True
            initprevsol = np.ones(nSt, dtype=bool)
            nrFail = np.zeros(sol.shape[0], dtype=int)
        tecarray[itm] = sol[:, 0]
        clockarray[itm] = sol[:, 1]
        if fit3rdorder:
            tec3rdarray[itm]=sol[:, 2]
    if returnResiduals:
        if fit3rdorder:
            return (tecarray, clockarray, residualarray,tec3rdarray)
        else:
            return (tecarray, clockarray, residualarray)
    if fit3rdorder:
        return (tecarray, clockarray,tec3rdarray)
    return (tecarray, clockarray)

def getClockTECFitStation(
    ph,
    freq,
    stationname,
    initSol=[],
    returnResiduals=True,
    chi2cut=1e8,
    fit3rdorder=False,
    double_search_space=False
    ):
    '''get the c/t separation per station'''
    nT = ph.shape[0]
    nF = freq.shape[0]
    data = ph
    tecarrayst = np.zeros((nT,), dtype=np.float32)
    clockarrayst = np.zeros((nT,), dtype=np.float32)

    if returnResiduals:
        residualarrayst = np.zeros((nT, nF), dtype=np.float32)
    if fit3rdorder:
        tec3rdarrayst= np.zeros((nT,), dtype=np.float32)
    A = np.ones((nF, 2+fit3rdorder), dtype=np.float)
    A[:, 1] = freq * 2 * np.pi * 1e-9
    A[:, 0] = -8.44797245e9 / freq
    if fit3rdorder:
        A[:, 2] = -1e21 / freq**3
    steps = np.ma.dot(np.ma.dot(np.linalg.inv(np.ma.dot(A.T, A)), A.T), 2 * np.pi * np.ones((freq.shape[0], ), dtype=np.float))
    succes=False
    initprevsol=False
    nrFail=0
    sol = np.zeros((2+fit3rdorder), dtype=np.float)
    prevsol = np.zeros_like(sol)
    n3rd=0
    for itm in range(nT):
        datatmp=np.ma.copy(data[itm, :])
        if itm == 0 or not succes:
            if itm == 0 or not initprevsol:
                if hasattr(initSol, '__len__') and len(initSol) > 0 :
                    sol[:initSol.shape[0]]=initSol
                    ndt=1
                    ndtec=1
                    if fit3rdorder:
                        n3rd=1
                else:
                    sol[:]=0
                    if fit3rdorder:
                        n3rd=200
                    if 'CS' in stationname:
                        ndt=4
                        ndtec=10
                        if 'LBA' in stationname:
                            ndt=2
                            ndtec=40
                    else:
                        if 'RS' in stationname:
                            ndt=200
                            ndtec=80
                        else:
                            # large TEC variation for EU stations
                            ndt=200
                            ndtec=160
                        if 'LBA' in stationname:
                            ndt=60
                            # no init clock possible due to large TEC effect
                            #stepsize of dtec is small
                            ndtec=320
            else:
                # further steps with non success
                sol[:] = prevsol[:]
                ndtec=min(nrFail+1,100)
                if not 'CS' in stationname:
                    ndt=min(nrFail+1,200)
                else:
                    ndt=min(nrFail+1,4)
                if fit3rdorder:
                    n3rd=min(nrFail+1,200)
            datatmpist = datatmp[:]
            if datatmpist.count() / float(nF) > 0.5:
                # do brutforce and update data, unwrp pdata,update flags
                #logging.debug("Getting init par for time %d:station %s ntec %d ndt %d n3rd %d"%(itm,stationname,ndtec,ndt,n3rd)+str(sol))
                try:
                    par,datatmp[:] = getInitPar(datatmpist, freq,nrTEC=ndtec*(1+double_search_space),nrClock=ndt*(1+double_search_space),nrthird=n3rd*(1+double_search_space),initsol=sol[:])
                    sol[:] = par[:]
                except np.linalg.LinAlgError:
                    logging.debug("Init parmaters failed   t=%d st=%s flags=%d" % (itm,stationname,datatmpist.count()) + str(sol[:]))
                    sol[:] = [-10.,]*sol.shape[0]
                    
                #logging.debug("Getting init par for station %d:%s "%(itm,stationname)+str(sol))

        #now do the real fitting
        datatmpist=datatmp[:]
        if datatmpist.count() / float(nF) < 0.5:
            logging.debug("Too many data points flagged 2nd  t=%d st=%s flags=%d" % (itm,stationname,datatmpist.count()) + str(sol[:]))
            sol[:] = [-10.,]*sol.shape[0]
        else:
            try:
                fitdata=np.dot(sol,A.T)
                datatmpist=unwrapPhases(datatmpist,freq,fitdata)
                mymask=datatmpist.mask
                maskedfreq=np.ma.array(freq,mask=mymask)
                A2=np.ma.array(A,mask=np.tile(mymask,(A.shape[1],1)).T)
                if datatmpist.count() / float(nF) < 0.5:
                    logging.debug("Too many data points flagged 3rd t=%d st=%s flags=%d" % (itm,stationname,datatmpist.count()) + str(sol[:]))
                    sol[:] = [-10.,]*sol.shape[0]
                else:
                    sol[:] = np.ma.dot(np.linalg.inv(np.ma.dot(A2.T, A2)), np.ma.dot(A2.T, datatmpist)).T
                if initprevsol and np.abs((sol[1]-prevsol[1])/steps[1])>0.5 and (np.abs((sol[1]-prevsol[1])/steps[1])>0.75 or np.abs(np.sum((sol-prevsol)/steps,axis=-1))>0.5*A2.shape[0]):
                    #logging.debug("removing jumps, par for station %d , itm %d"%(ist,itm)+str(sol[ist])+str(prevsol[ist])+str(steps))
                    sol[:]-=np.round((sol[1]-prevsol[1])/steps[1])*steps
            except np.linalg.LinAlgError:
                logging.debug("Fit failed   t=%d st=%s flags=%d" % (itm,stationname,datatmpist.count()) + str(sol[:]))
                sol[:] = [-10.,]*sol.shape[0]
               
        # calculate chi2 per station

        residual = data[itm] - np.dot(A, sol)
        residual = np.ma.remainder(residual + np.pi, 2 * np.pi) - np.pi
        chi2 = np.ma.sum(np.square(np.degrees(residual)), axis=0) / nF

        if returnResiduals:
            residualarrayst[itm] = residual

        chi2select = (chi2 > chi2cut) or (sol[0] < -5) # select bad points
        chi2select = chi2select or initprevsol*np.sum(np.abs((sol-prevsol)/steps),axis=-1)>(0.3*sol.shape[0]*(1+nrFail)) #also discard data where there is a "half" jump for any parameter wrst the previous solution (if previous solution exists). Multiply with number of fails, since after a large number of fails there is no clear match with the previous solution expected anyway...
        if not initprevsol:
            prevsol= sol[:]
        if  chi2select:
            logging.debug('High chi2 of fit, itm: %d  ' % (itm) + 'station:' + stationname)
            succes = False
            nrFail += 1
        else:
            prevsol = 0.5 * prevsol + 0.5 * np.copy(sol)
            succes = True
            initprevsol = True
            nrFail = 0
        tecarrayst[itm] = sol[0]
        clockarrayst[itm] = sol[1]
        if fit3rdorder:
            tec3rdarrayst[itm]=sol[2]
    if returnResiduals:
        if fit3rdorder:
            return (tecarrayst, clockarrayst, residualarrayst,tec3rdarrayst)
        else:
            return (tecarrayst, clockarrayst, residualarrayst)
    if fit3rdorder:
        return (tecarrayst, clockarrayst,tec3rdarrayst)
    return (tecarrayst, clockarrayst)


def getPhaseWrapBase(freqs):
    """
    freqs: frequency grid of the data
    return the step size from a local minima (2pi phase wrap) to the others [0]: TEC, [1]: clock
    """

    nF = freqs.shape[0]
    A = np.zeros((nF, 2), dtype=np.float)
    A[:, 1] = freqs * 2 * np.pi * 1e-9
    A[:, 0] = -8.44797245e9 / freqs
    steps = np.dot(np.dot(np.linalg.inv(np.dot(A.T, A)), A.T), 2 * np.pi * np.ones((nF, ), dtype=np.float))
    basef = np.dot(A, steps) - 2 * np.pi
    return (basef, steps)


def getResidualPhaseWraps(avgResiduals, freqs):
    flags = np.average(avgResiduals.mask,axis=1)>0.5
    nSt = avgResiduals.shape[1]
    nF = freqs.shape[0]
    wraps = np.zeros((nSt, ), dtype=np.float)
    tmpflags = flags
    tmpfreqs = freqs[np.logical_not(tmpflags)]
    steps=[0,0]
    if np.ma.count(tmpfreqs) < 3:
        logging.debug('Cannot unwrap, too many channels flagged')
        return (wraps,steps)
    (tmpbasef, steps) = getPhaseWrapBase(tmpfreqs)
    basef = np.zeros(freqs.shape)
    basef[np.logical_not(tmpflags)] = tmpbasef
    basef = basef.reshape((-1, 1))
    data = avgResiduals[:, :]
    if has_fitting:
        wraps = fitting.fit(data, basef, wraps, flags).flatten()
    return (wraps, steps)


def correctWrapsFromResiduals(residualarray,flags,freq):
    '''corrects solution jumps due to 2 pi phasewraps based on the average residuals'''
    resflags = np.logical_or(flags[:, np.newaxis], residualarray == 0)
    maskedresiduals = np.ma.array(residualarray, mask=resflags)
    # avgResiduals=np.average(residualarray,axis=0)
    avgResiduals = np.ma.average(maskedresiduals, axis=0)
    (wraps, steps) = getResidualPhaseWraps(avgResiduals, freq)  # fitting of the wraps from the time-avg residuals
    # step[0] is the step in TEC corresponding to 1 phase wrap and step[1] is in ns (clock)
    wraps = np.round(wraps - wraps[0])  # reference to station 0
    logging.debug('Wraps from residuals: ' + str(wraps))
    return (wraps, steps)


def correctWraps(
    tecarray,
    residualarray,
    freq,
    pos,
    ):
    '''corrects solution jumps due to 2 pi phasewraps based on spatial correlations of averaged TEC solutions. Also returns average constant phase  offset per station. International stations should be excluded from this'''
    nT = tecarray.shape[0]
    nSt = tecarray.shape[1]
    flags = tecarray < -5

    (wraps, steps) = correctWrapsFromResiduals(residualarray,flags,freq)
    lats = np.degrees(np.arctan2(pos[:, 2], np.sqrt(pos[:, 0] * pos[:, 0] + pos[:, 1] * pos[:, 1])))
    lats -= lats[0]
    lons = np.degrees(np.arctan2(pos[:, 1], pos[:, 0]))
    lons -= lons[0]
    lonlat = np.concatenate((lons, lats)).reshape((2, ) + lons.shape)
    # refine (is it needed/helpful for LBA? the offsets seem to make sense for LBA)
    myselect=np.sqrt(np.sum(lonlat**2,axis=0))>0.5
    

    for nr_iter in range(2):
        # recreate best TEC at the moment
        TEC = tecarray - tecarray[:, [0]] + steps[0] * (np.round(wraps) - np.round(wraps[0]))
        #TEC = np.ma.array(TEC, mask=flags)
        TEC = np.ma.array(TEC, mask=np.logical_or(flags,myselect[np.newaxis]))
        # fit 2d linear TEC screen over stations
        slope = np.ma.dot(np.linalg.inv(np.dot(lonlat, lonlat.T)), np.ma.dot(lonlat, TEC.T))
        # flag bad time steps maybe because TEC screen is not linear
        chi2 = np.ma.sum(np.square(TEC - np.ma.dot(lonlat.T, slope).T), axis=1) / nSt
        chi2select = chi2 < np.ma.average(chi2)
        chi2select = chi2 < np.ma.average(chi2[chi2select])
        # calculate offset per station wrt time-averaged TEC screen
        offsets = -1 * np.ma.average(TEC[chi2select] - np.ma.dot(slope.T, lonlat)[chi2select], axis=0) * 2. * np.pi / steps[0]
        remainingwraps = np.round(offsets / (2 * np.pi))  # -np.round(wraps[stationIndices])
        offsets[myselect]=0
        remainingwraps[myselect]=0
        logging.debug('Offsets: ' + str(offsets))
        logging.debug('AvgTEC: ' + str(np.ma.average(TEC[chi2select], axis=0)))
        logging.debug('Remaining: ' + str(remainingwraps))
        wraps += remainingwraps
        # TODO: remove also the offset before the second cycle
        if np.sum(np.absolute(remainingwraps)) == 0:
            break
    return (offsets, wraps, steps)

def get_first_good(data,axis=1,check=lambda x:np.logical_and(x!=-10,x !=0)):
    a=check(data).nonzero()
    ndata=[]
    for i in range(data.shape[axis]):
        if np.any(a[axis]==i):
            idx=np.where(a[axis]==i)[0][0]
        else:
            idx=0
        ndata.append(np.take(data,a[abs(1-axis)][idx],axis=abs(1-axis))[i])
    return np.array(ndata)


def merge_ct_args(args):
    '''hel[per function for multiprocessing'''
    return getClockTECFitStation(*args)


def doFit(
    phases,
    mask,
    freqs,
    stations,
    station_positions,
    axes,
    refstIdx='superterp',
    flagBadChannels=True,
    flagcut=1.5,
    chi2cut=30000.,
    removePhaseWraps=True,
    combine_pol=False,
    fit3rdorder=False,
    circular=False,
    initSol=[],
    initoffsets=[],
    n_proc=10
    ):
    # make sure order of axes is as expected
    stidx = axes.index('ant')
    freqidx = axes.index('freq')
    timeidx = axes.index('time')
    try:
        polidx = axes.index('pol')
        data = ma.array(phases, mask=mask).transpose((timeidx, freqidx, stidx, polidx))
        npol = data.shape[3]
    except:
        data = ma.array(phases, mask=mask).transpose((timeidx, freqidx, stidx))
        data = data.reshape(data.shape+(1,))
        npol = 1

    nT = data.shape[0]
    nF = data.shape[1]
    nSt = data.shape[2]

    if npol == 4:
        data = data[:, :, :, (0, 3)]
        npol = 2
    if refstIdx == 'superterp':
        superterpstations = [i for i in stations if i[:5] in [
            'CS002',
            'CS003',
            'CS004',
            'CS005',
            'CS006',
            'CS007'
        ]]
        refstIdx = [i for (i, j) in enumerate(stations) if j in superterpstations]
        if len(refstIdx)==0:
            refstIdx=0
    if not hasattr(refstIdx, '__len__'):
        refstIdx = [refstIdx]
    # get phases from reference stations
    refdata = np.ma.sum(np.cos(data[:, :, refstIdx, :]) + 1.j * np.sin(data[:, :, refstIdx, :]), axis=2)
    refdata = np.ma.arctan2(np.imag(refdata), np.real(refdata))
    # unwrap around mean
    mymean = np.ma.average(refdata, axis=0)
    refdata = np.ma.remainder(refdata - mymean[np.newaxis] + np.pi, 2 * np.pi) + mymean[np.newaxis] - np.pi
    data -= refdata[:, :, np.newaxis]
    # flag bad channels - test if really nedded if data already flagged
    indices = np.arange(nF)

    if flagBadChannels:
        mymask=np.zeros((nF), dtype=np.bool)
        for nr_iter in range(2):
            rms = np.ma.std(np.ma.std(refdata, axis=0), axis=1)
            freqselect = rms < flagcut * np.average(rms)
            logging.debug('Iter %d: flagging %d channels' % (nr_iter, np.sum(np.logical_not(np.array(freqselect)))))
            #freqs = freqs[freqselect]
            #data = data[:, freqselect]
            mymask=np.ma.logical_or(mymask,~freqselect)
            #refdata = refdata[:, freqselect]
            refdata.mask=np.ma.logical_or(refdata.mask,mymask[np.newaxis,:,np.newaxis])
            logging.debug('Flagging: ' + str(indices[np.logical_not(freqselect)]))
            #indices = indices[freqselect]
        mask=data.mask
        data.mask=np.logical_or(mask,mymask[np.newaxis,:,np.newaxis,np.newaxis])

    # Remove completely flagged stations
    selectstations = [st for stationIndices, st in enumerate(stations) if data[:, :, stationIndices].mask.all() != True]
    logging.debug('%d selected stations: ' % len(selectstations) + str(selectstations))
    stationIndices = np.array([idxst in selectstations for idxst in stations])
    data = data[:, :, stationIndices]
    stations_old = stations[:] # keep a copy for putting flagged stations back at the end
    stations = stations[stationIndices]

    station_positions = station_positions[stationIndices]
    RSstations = [i for (i, j) in enumerate(stations) if 'RS' in j]
    CSstations = [i for (i, j) in enumerate(stations) if 'CS' in j]
    otherstations = [i for (i, j) in enumerate(stations) if not 'RS' in j and not 'CS' in j]
    logging.debug('Station indices: ' + str(stationIndices) + ' RS ' + str(RSstations))
    nSt = data.shape[2]
    # combine polarizationsif requested - needed in HBA
    if combine_pol:
        if npol == 2 and not circular:
            cdata = np.ma.cos(data) + 1.j * np.ma.sin(data)
            data = np.ma.sum(cdata, axis=3).reshape((nT, nF, nSt, 1))
            data = np.ma.arctan2(np.imag(data), np.real(data))  # np.angle doesnot yet return masked array!!
            npol = 1
        if circular:
            data=np.ma.sum(data, axis=3).reshape((nT, nF, nSt, 1))
            npol=1
    # guess clock, remove from data
    # not in LBA because TEC dominant
    if not 'LBA' in stations[0] and len(initSol) < 1:
        initclock = getInitClock(data[int(nT / 2):int(nT / 2 + 100)][:, :, RSstations + otherstations], freqs)  # only on a few timestamps
        logging.debug('Initial clocks: ' + str(initclock[1]))
        # init CS clocks to 0
        # logging.debug("data before init clock" + str(data[nT/2,:,-1]))
        data[:, :, RSstations + otherstations] = data[:, :, RSstations + otherstations] - freqs[np.newaxis, :, np.newaxis, np.newaxis] * initclock[1][np.newaxis, np.newaxis, :] * 1e-9 * 2 * np.pi
        # logging.debug("clock correction" + str(np.remainder(freqs*initclock[1][-1]*-1e-9*2*np.pi+np.pi,2*np.pi)-np.pi))
        # logging.debug("data after init clock" + str(np.remainder(data[nT/2,:,-1]+np.pi,2*np.pi)-np.pi))
    offset = np.zeros((nSt, npol), dtype=np.float32)
    if len(initoffsets)>0: # Check if initoffsets is not empty
        offset = initoffsets
        data[:, :, :, :] += offset[:][np.newaxis, np.newaxis]
    # initialize arrays
    clock = np.zeros((nT, nSt, npol), dtype=np.float32)
    tec = np.zeros((nT, nSt, npol), dtype=np.float32)
    if fit3rdorder:
        tec3rd = np.zeros((nT, nSt, npol), dtype=np.float32)

    # better not to use fitoffset
    tecarray = np.zeros((nT, nSt), dtype=np.float32)
    clockarray = np.zeros((nT, nSt), dtype=np.float32)

    residualarray = np.zeros((nT, nF, nSt), dtype=np.float32)
    if fit3rdorder:
        tec3rdarray= np.zeros((nT, nSt), dtype=np.float32)
    for pol in range(npol):
        # get a good guesss without offset
        # logging.debug("sending masked data "+str(data[:,:,:,pol].count()))
        initialchi2cut = chi2cut  # user defined
        if removePhaseWraps:
            initialchi2cut = 30000.  # this number is quite arbitrary
        if fit3rdorder:
              poolargs=[]
              for ist in range(nSt):
                  poolargs.append((np.ma.copy(data[:, :, ist, pol]),
                                   freqs,
                                   stations[ist],
                                   [],
                                   True,
                                   initialchi2cut,
                                   True,
                                   circular and combine_pol
                               ))
              p=Pool(processes=n_proc)
              result=p.map(merge_ct_args,poolargs)
              for ist,tc in enumerate(result):
                  tecarray[:, ist], clockarray[:, ist],residualarray[:,:,ist],tec3rdarray[:,ist] = tc
        else:
              poolargs=[]
              for ist in range(nSt):
                  poolargs.append((np.ma.copy(data[:, :, ist, pol]),
                                   freqs,
                                   stations[ist],
                                   [],
                                   True,
                                   initialchi2cut,
                                   False,
                                   circular and combine_pol
                               ))
              p=Pool(processes=n_proc)
              result=p.map(merge_ct_args,poolargs)
              for ist,tc in enumerate(result):
                  tecarray[:, ist], clockarray[:, ist],residualarray[:,:,ist] = tc
        if removePhaseWraps:
            # correctfrist times only,try to make init correct ?
            #corrects wraps based on spatial correlation (averaged in time), only works for long time observations, not testted for LBA
            (offset[:, pol], wraps, steps) = correctWraps(tecarray, residualarray, freqs, station_positions)
        else:
            #always correct for wraps based on average residuals
            wraps, steps = correctWrapsFromResiduals(residualarray, tecarray<-5,freqs)
        #maximum allowed 2pi phase wrap is +-1
        #wraps = np.minimum(wraps,2)
        #wraps = np.maximum(wraps,-2)
        logging.debug('Residual iter 1, pol %d: ' % pol + str(residualarray[0, 0]))
        logging.debug('TEC iter 1, pol %d: ' % pol + str(tecarray[0]))
        logging.debug('Clock iter 1, pol %d: ' % pol + str(clockarray[0]))
        logging.debug('Wraps: ' + str(wraps))
        logging.debug('Offsets: ' + str(offset[:, pol]))
        # remove completely initialoffset?
        if len(initoffsets)>0: # Check if initoffsets is not empty
            offset[:, pol] -= initoffsets[:, pol]
        data[:, :, :, pol] += offset[:, pol][np.newaxis, np.newaxis]
        # remove fitoffset
        if removePhaseWraps:
            initsol = np.zeros((nSt, 2), dtype=np.float32)
            initsol[:, 0] = get_first_good(tecarray[:, :]) + wraps * steps[0]
            initsol[:, 1] = get_first_good(clockarray[:, :]) + wraps * steps[1]
            logging.debug('Initsol TEC, pol %d: ' % pol + str(initsol[:, 0]))
            logging.debug('Initsol clock, pol %d: ' % pol + str(initsol[:, 1]))
            # is it needed to redo the fitting? this is the time bottleneck
            if fit3rdorder:
              poolargs=[]
              for ist in range(nSt):
                  poolargs.append((np.ma.copy(data[:, :, ist, pol]),
                                   freqs,
                                   stations[ist],
                                   initsol[ist],
                                   False,
                                   chi2cut,
                                   True,
                                   circular and combine_pol
                               ))
              p=Pool(processes=n_proc)
              result=p.map(merge_ct_args,poolargs)
              for ist,tc in enumerate(result):
                  tec[:, ist, pol], clock[:, ist, pol],tec3rd[:,ist,pol] = tc
            else:
              poolargs=[]
              for ist in range(nSt):
                  poolargs.append((np.ma.copy(data[:, :, ist, pol]),
                                   freqs,
                                   stations[ist],
                                   initsol[ist],
                                   False,
                                   chi2cut,
                                   False,
                                   circular and combine_pol
                               ))
              p=Pool(processes=n_proc)
              result=p.map(merge_ct_args,poolargs)
              for ist,tc in enumerate(result):
                  tec[:, ist, pol], clock[:, ist, pol] = tc

        else:
            tec[:, :, pol] = tecarray[:, :]+ wraps * steps[0]
            clock[:, :, pol] = clockarray[:, :]+ wraps * steps[1]
            if fit3rdorder:
              tec3rd[:, :, pol]  = tec3rdarray[:, :]+ wraps * steps[2]
        logging.debug('TEC iter 2, pol %d: ' % pol + str(tec[0, :, pol]))
        logging.debug('Clock iter 2, pol %d: ' % pol + str(clock[0, :, pol]))
    if not 'LBA' in stations[0] and len(initSol) < 1:
        clock[:, RSstations + otherstations] += initclock[1][np.newaxis, :, :]
    if combine_pol and circular:
        clock/=2;tec/=2;offset/=2

    # put flagged stations back
    for idx, wasUsed in enumerate(stationIndices):
        if not wasUsed:
            logging.debug('Adding completely flagged station '+str(idx))
            clock = np.insert(clock,idx,0,axis=1) # [time:ant:pol]
            tec = np.insert(tec,idx,-5,axis=1) # [time:ant:pol]
            offset = np.insert(offset,idx,0,axis=0) # [ant:pol]
            if fit3rdorder: tec3rd = np.insert(tec3rd,idx,0,axis=1) # [time:ant:pol]

#    station = station_old
    if fit3rdorder:
        return (clock, tec, offset, tec3rd)
    else:
        return (clock, tec, offset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading FLAG module.')

def _run_parser(soltab, parser, step):
    axesToFlag = parser.getarraystr( step, 'axesToFlag') # no default
    order = parser.getarrayint( step, 'order') # no default
    maxCycles = parser.getint( step, 'maxCycles', 5)
    maxRms = parser.getfloat( step, 'maxRms', 5.)
    maxRmsNoise = parser.getfloat( step, 'maxRmsNoise', 0.)
    fixRms = parser.getfloat( step, 'fixRms', 0.)
    fixRmsNoise = parser.getfloat( step, 'fixRmsNoise', 0.)
    windowNoise = parser.getint( step, 'windowNoise', 11)
    replace = parser.getbool( step, 'replace', False)
    preflagzeros = parser.getbool( step, 'preflagzeros', False)
    mode = parser.getstr( step, 'mode', 'smooth')
    refAnt = parser.getstr( step, 'refAnt', '')
    ncpu = parser.getint( '_global', 'ncpu', 0)

    parser.checkSpelling( step, soltab, ['axesToFlag', 'order', 'maxCycles', 'maxRms', 'maxRmsNoise', 'fixRms', 'fixRmsNoise', 'windowNoise', 'replace', 'preflagzeros', 'mode', 'refAnt'])
    return run( soltab, axesToFlag, order, maxCycles, maxRms, maxRmsNoise, fixRms, fixRmsNoise, windowNoise, replace, preflagzeros, mode, refAnt, ncpu )


def _flag(vals, weights, coord, solType, order, mode, preflagzeros, maxCycles, maxRms, maxRmsNoise, windowNoise, fixRms, fixRmsNoise, replace, axesToFlag, selection, outQueue):

    import numpy as np
    import itertools
    from scipy.ndimage import generic_filter
    import scipy.interpolate

    def rolling_rms(a, window):
        """
        Return the rms for each element of the array calculated using the element inside a window,
        edges are mirrored
        """
        assert window % 2 == 1 # window must be odd
        a = np.pad(a, window/2, mode='reflect')
        shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
        strides = a.strides + (a.strides[-1],)
        return np.sqrt(np.var(np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides), -1))

    def polyfit(x=None, y=None, z=None, w=None, order=None):
        """Two-dimensional polynomial fit.

        References:
            http://stackoverflow.com/questions/7997152/
            python-3d-polynomial-surface-fit-order-dependent/7997925#7997925

        x: array of coordinates
        y: array of coordinates (2d only)
        z: values of these coordinates (1d array as x and y)
        w: weights
        order: int for 1d, tuple of 2 values (x-order, y-order) for 2d

        Return: matrix of coefficents, for the 2d is ok to be fed into polynomial.polyval2d()
        """
        assert z.shape == w.shape
        if y is None:
            # note that np.polynomial.polynomial.polyfit want sigmas not variances
            m = np.polyfit(x, z, order, w=np.sqrt(w))

        else:
            from numpy.polynomial import polynomial
            x, y = np.meshgrid(np.asarray(x),np.asarray(y), indexing='ij')
            # TODO: unset x,y,z value if flagged
            z = z[(w != 0)]
            x = x[(w != 0)]
            y = y[(w != 0)]
            vander = polynomial.polyvander2d(x, y, order)
            vander = vander.reshape((-1,vander.shape[-1]))
            z = z.reshape((vander.shape[0],))
            m = np.linalg.lstsq(vander, z)[0]
            order = np.asarray(order)
            m = m.reshape(order+1) # matrix of coefficents

        return m

    def polyval(x=None, y=None, m=None):
        """Values to two-dimensional polynomial fit. Based uppon code
            provided by Joe Kington.
        """
        if y is None:
            poly = np.poly1d(m)
            return poly(x)

        else:
            x, y = np.meshgrid(x,y, indexing='ij')
            return np.polynomial.polynomial.polyval2d(x, y, m)


    def outlier_rej(vals, weights, axes, order=5, mode='smooth', max_ncycles=3, max_rms=3., max_rms_noise=0., window_noise=11., fix_rms=0., fix_rms_noise=0., replace=False):
        """
        Reject outliers using a running median
        val = the array (avg must be 0)
        weights = the weights to convert into flags
        axes = array with axes values (1d or 2d)
        order = "see polyfit()"
        max_ncycles = maximum number of cycles
        max_rms = number of rms times for outlier flagging
        max_rms_noise = cut on the rms of the rmss
        window_noise = window used to calculate the rmss to detect noise
        replace = instead of flag it, replace the data point with the smoothed one

        return: flags array and final rms
        """

        # renormalize axes to have decent numbers
        if len(axes) == 1:
            axes[0] -= axes[0][0]
            axes[0] /= axes[0][1] - axes[0][0]
        elif len(axes) == 2:
            axes[0] -= axes[0][0]
            axes[0] /= (axes[0][1]-axes[0][0])
            axes[1] -= axes[1][0]
            axes[1] /= (axes[1][1]-axes[1][0])
        else:
            logging.error('FLAG operation can flag only along 1 or 2 axes. Given axes: '+str(axesToFlag))
            return

        # test artificial data
        #axes[0] = np.array(range(len(axes[0])))*24414.0625
        #axes[1] = np.array(range(len(axes[1])))*5
        #vals = np.empty( shape=[len(axes[0]), len(axes[1])] )
        #for i, xi in enumerate(axes[0]):
        #    for j, yj in enumerate(axes[1]):
        #        vals[i,j] = 10*xi + yj**2
        #weights = np.ones_like(vals)

        if replace:
            orig_weights = np.copy(weights)

        for i in range(max_ncycles):

            # all is flagged? break
            if (weights == 0).all():
                rms = 0.
                break

            if mode == 'smooth':
                vals_smooth = np.copy(vals)
                np.putmask(vals_smooth, weights==0, np.nan)
                # speedup: if all data are used then just do a median and don't call the filter
                if all(o == 0 for o in order):
                    vals_smooth = np.ones(vals_smooth.shape)*np.nanmedian(vals_smooth)
                else:
                    for i, o in enumerate(order): 
                        if o == 0: order[i] = vals_smooth.shape[i]
                    vals_smooth = generic_filter(vals_smooth, np.nanmedian, size=order, mode='constant', cval=np.nan)
                vals_detrend = vals - vals_smooth
            # TODO: should be rolling
            elif mode == 'poly':
                # get polynomia and values
                if len(axes) == 1:
                    fit_sol = polyfit(axes[0], z=vals, w=weights, order=order)
                    vals_detrend = vals - polyval(axes[0], m=fit_sol)
                elif len(axes) == 2:
                    fit_sol = polyfit(axes[0], axes[1], z=vals, w=weights, order=order)
                    vals_smooth = polyval(axes[0], axes[1], m=fit_sol)
                    vals_detrend = vals - vals_smooth
            # TODO: should be rolling
            elif mode == 'spline':
                # get spline
                if len(axes) == 1:
                    spline = scipy.interpolate.UnivariateSpline(axes[0], y=vals, w=weights, k=order[0])
                    vals_detrend = vals - spline(axes[0])
                elif len(axes) == 2:
                    x, y = np.meshgrid(axes[0], axes[1], indexing='ij')
                    # spline doesn't like w=0
                    z = vals[(weights != 0)].flatten()
                    x = x[(weights != 0)].flatten()
                    y = y[(weights != 0)].flatten()
                    w = weights[(weights != 0)].flatten()
                    spline = scipy.interpolate.SmoothBivariateSpline(x, y, z, w, kx=order[0], ky=order[1])
                    vals_smooth = spline(axes[0], axes[1])
                    vals_detrend = vals - vals_smooth

            # remove outliers
            if max_rms > 0 or fix_rms > 0:
                # median calc https://en.wikipedia.org/wiki/Median_absolute_deviation
                rms =  1.4826 * np.nanmedian( np.abs(vals_detrend[(weights != 0)]) )
                if np.isnan(rms): weights[:] = 0
                elif fix_rms > 0:
                    flags = abs(vals_detrend) > fix_rms
                    weights[ flags ] = 0
                else:
                    flags = abs(vals_detrend) > max_rms * rms
                    weights[ flags ] = 0

            # remove noisy regions of data
            if max_rms_noise > 0 or fix_rms_noise > 0:
                rmses = rolling_rms(vals_detrend, window_noise)
                rms =  1.4826 * np.nanmedian( abs(rmses) )

                # rejection
                if fix_rms_noise > 0:
                    flags = rmses > fix_rms_noise
                else:
                    flags = rmses > (max_rms_noise * rms)
                weights[ flags ] = 0

            # all is flagged? break
            if (weights == 0).all():
                rms == 0.
                break

            # no flags? break
            if (flags == False).all():
                break

        # replace (outlier) flagged values with smoothed ones
        if replace:
            logging.debug('Replacing %.2f%% of the data.' % np.sum(orig_weights != weights))
            vals[np.where(orig_weights != weights)] = vals_smooth[np.where(orig_weights != weights)]
            weights = orig_weights

        # plot 1d
        plot = False
        if plot:
            import matplotlib as mpl
            mpl.use("Agg")
            import matplotlib.pyplot as plt
            plt.plot(axes[1][weights[0] == 0], vals[0][weights[0] == 0], 'ro')
            plt.plot(axes[1], vals[0], 'k.')
            plt.plot(axes[1], vals_smooth[0], 'r.')
            plt.plot(axes[1], vals_detrend[0], 'g.')
            plt.savefig('test.png')
            sys.exit(1)

        return weights, vals, rms


    def percentFlagged(w):
        return 100.*(weights.size-np.count_nonzero(weights))/float(weights.size)
    ########################################

    # check if everything flagged
    if (weights == 0).all() == True:
        logging.debug('Percentage of data flagged/replaced (%s): already completely flagged' % (removeKeys(coord, axesToFlag)))
        outQueue.put([vals, weights, selection])
        return

    if preflagzeros:
        if solType == 'amplitude': np.putmask(weights, vals == 1, 0)
        else: np.putmask(weights, vals == 0, 0)

    flagCoord = []
    for axisToFlag in axesToFlag:
        flagCoord.append(coord[axisToFlag])

    initPercentFlag = percentFlagged(weights)

    # works in phase-space (assume no wraps), remove just the mean to prevent problems if the phase is constantly around +/-pi
    if solType == 'phase' or solType == 'scalarphase' or solType == 'rotation':
        # remove mean of vals
        mean = np.angle( np.sum( weights.flatten() * np.exp(1j*vals.flatten()) ) / ( vals.flatten().size * sum(weights.flatten()) ) )
        logging.debug('Working in phase-space, remove angular mean '+str(mean)+'.')
        vals = normalize_phase(vals - mean)
        weights, vals, rms = outlier_rej(vals, weights, flagCoord, order, mode, maxCycles, maxRms, maxRmsNoise, windowNoise, fixRms, fixRmsNoise, replace)
        vals = normalize_phase(vals + mean)

    elif solType == 'amplitude':
        vals_good = (vals>0)
        vals[vals_good] = np.log10(vals[vals_good])
        weights, vals, rms = outlier_rej(vals, weights, flagCoord, order, mode, maxCycles, maxRms, maxRmsNoise, windowNoise, fixRmsNoise, replace)
        vals[vals_good] = 10**vals[vals_good]

    else:
        weights, vals, rms = outlier_rej(vals, weights, flagCoord, order, mode, maxCycles, maxRms, maxRmsNoise, windowNoise, fixRms, fixRmsNoise, replace)

    if percentFlagged(weights) == initPercentFlag:
        logging.debug('Percentage of data flagged/replaced (%s): %.3f -> None' % ((removeKeys(coord, axesToFlag), initPercentFlag)))
    else:
        logging.debug('Percentage of data flagged/replaced (%s): %.3f -> %.3f %% (rms: %.5f)' \
            % ((removeKeys(coord, axesToFlag), initPercentFlag, percentFlagged(weights), rms)))

    outQueue.put([vals, weights, selection])
#    return vals, weights, selection


def run( soltab, axesToFlag, order, maxCycles=5, maxRms=5., maxRmsNoise=0., fixRms=0., fixRmsNoise=0., windowNoise=11, replace=False, preflagzeros=False, mode='smooth', refAnt='', ncpu=0 ):
    """
    This operation for LoSoTo implement a flagging procedure
    WEIGHT: compliant

    Parameters
    ----------
    axesToFlag : array of str
        Axes along which to smooth+find outlier (e.g. ['time', 'freq']), max 2 values.

    order : array of int
        Order of the function fitted during detrending. Array must have same size of axesToFlag. If mode=smooth these are the window of the running median (0=all axis).

    maxCycles : int, optional
        Max number of independent flagging cycles, by default 5.

    maxRms : float, optional
        Rms to clip outliers, by default 5.

    maxRmsNoise : float, optional
        Do a running rms and then flag those regions that have a rms higher than MaxRmsNoise*rms_of_rmses, by default 0 (ignored).

    fixRms : float, optional
        Instead of calculating rms use this value, by default 0 (ignored).

    fixRmsNoise : float, optional
        Instead of calculating rms of the rmses use this value (it will not be multiplied by the MaxRmsNoise), by default 0 (ignored).

    windowNoise : int, optional
        Window size for the running rms, by default 11.

    replace : bool, optional
        Replace bad values with the interpolated ones, instead of flagging them. By default False.

    preflagzeros : bool, optional
        Flag zeros/ones (bad solutions in BBS/DPPP). They should be flagged at import time. By default False.

    mode: str, optional
        Detrending/fitting algorithm: smooth / poly / spline. By default smooth.

    refAnt : str, optional
        Reference antenna, by default None.

    ncpu : int, optional
        Number of cpu to use, by default all available.
    """

    logging.info("Flag on soltab: "+soltab.name)

    # input check
    if refAnt == '':
        refAnt = None

    mode = mode.lower()
    if mode not in ['smooth', 'poly', 'spline']:
        logging.error('Mode must be one of smooth, poly, or spline')
        return 1

    if axesToFlag == []:
        logging.error("Please specify axis to flag. It must be a single one.")
        return 1

    for axisToFlag in axesToFlag:
        if axisToFlag not in soltab.getAxesNames():
            logging.error('Axis \"'+axisToFlag+'\" not found.')
            return 1

    if len(axesToFlag) != len(order) and (len(axesToFlag) != 1 or len(axesToFlag) != 2):
        logging.error("AxesToFlag and order must be both 1 or 2 values.")
        return 1

    if len(order) >= 2: order = list(order)

    # start processes for multi-thread
    mpm = multiprocManager(ncpu, _flag)

    # reorder axesToFlag as axes in the table
    axesToFlag_orig = axesToFlag
    axesToFlag = [coord for coord in soltab.getAxesNames() if coord in axesToFlag]
    if axesToFlag_orig != axesToFlag: order = order[::-1] # reverse order if we changed axesToFlag

    solType = soltab.getType()

    # fill the queue (note that sf and sw cannot be put into a queue since they have file references)
    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=axesToFlag, weight=True, refAnt=refAnt):
        mpm.put([vals, weights, coord, solType, order, mode, preflagzeros, maxCycles, maxRms, maxRmsNoise, windowNoise, fixRms, fixRmsNoise, replace, axesToFlag, selection])

    mpm.wait()

    for v, w, sel in mpm.get():
        if replace:
            # rewrite solutions (flagged values are overwritten)
            soltab.setValues(v, sel, weight=False)
        else:
            soltab.setValues(w, sel, weight=True)

    soltab.flush()
    soltab.addHistory('FLAG (over %s with %s sigma cut)' % (axesToFlag, maxRms))

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging
from losoto.lib_unwrap import unwrap, unwrap_2d

logging.debug('Loading PLOT module.')

def _run_parser(soltab, parser, step):
    axesInPlot = parser.getarraystr( step, 'axesInPlot' ) # no default
    axisInTable = parser.getstr( step, 'axisInTable', '' )
    axisInCol = parser.getstr( step, 'axisInCol', '' )
    axisDiff = parser.getstr( step, 'axisDiff', '' )
    NColFig = parser.getint( step, 'NColFig', 0 )
    figSize = parser.getarrayint( step, 'figSize', [0,0] )
    markerSize = parser.getint( step, 'markerSize', 2 )
    minmax = parser.getarrayfloat( step, 'minmax', [0.,0.] )
    log = parser.getstr( step, 'log', '' )
    plotFlag = parser.getbool( step, 'plotFlag', False )
    doUnwrap = parser.getbool( step, 'doUnwrap', False )
    refAnt = parser.getstr( step, 'refAnt', '' )
    refDir = parser.getstr( step, 'refDir', '' )
    soltabsToAdd = parser.getarraystr( step, 'soltabsToAdd', [] )
    makeAntPlot = parser.getbool( step, 'makeAntPlot', False )
    makeMovie = parser.getbool( step, 'makeMovie', False )
    prefix = parser.getstr( step, 'prefix', '' )
    ncpu = parser.getint( '_global', 'ncpu', 0 )

    parser.checkSpelling( step, soltab, ['axesInPlot', 'axisInTable', 'axisInCol', 'axisDiff', 'NColFig', 'figSize', 'markerSize', 'minmax', 'log', \
               'plotFlag', 'doUnwrap', 'refAnt', 'refDir', 'soltabsToAdd', 'makeAntPlot', 'makeMovie', 'prefix'])
    return run(soltab, axesInPlot, axisInTable, axisInCol, axisDiff, NColFig, figSize, markerSize, minmax, log, \
               plotFlag, doUnwrap, refAnt, refDir, soltabsToAdd, makeAntPlot, makeMovie, prefix, ncpu)

def _plot(Nplots, NColFig, figSize, markerSize, cmesh, axesInPlot, axisInTable, xvals, yvals, xlabelunit, ylabelunit, datatype, filename, titles, log, dataCube, minZ, maxZ, plotFlag, makeMovie, antCoords, outQueue):
 
    import sys
    from itertools import cycle, chain
    if not 'matplotlib' in sys.modules:
        import matplotlib as mpl
        mpl.rcParams['xtick.labelsize'] = 20
        mpl.rcParams['font.size'] = 20
        mpl.use("Agg")
    import matplotlib.pyplot as plt # after setting the backend

    # find common min and max if not set
    flat = dataCube.filled(np.nan).flatten()
    if np.isnan(flat).all() or np.all(flat==0):
        minZ=-0.1; maxZ=0.1
    elif minZ == 0 and maxZ == 0:
        if datatype == 'phase':
            minZ = np.nanmin(flat)
            maxZ = np.nanmax(flat)
        elif datatype == 'amplitude' and len(axesInPlot) > 1:
            flat[np.isnan(flat)] = np.nanmedian(flat) # get rid of nans (problem in "<" below)
            maxZ = np.nanmedian( flat ) + 3*np.nanstd( flat[ (flat / np.nanmedian(flat) ) < 100  ] )
            maxZ = np.nanmin( [np.nanmax( flat ), maxZ] )
            minZ = np.nanmin( flat )
        else:
            minZ = np.nanmin(flat)
            maxZ = np.nanmax(flat)
 
        # prevent same min/max (still a problem at 0)
        if minZ == maxZ:
            if minZ > 0:
                minZ *= 0.99
                maxZ *= 1.01
            else:
                minZ *= 1.01
                maxZ *= 0.99
            
        # add some space for clock plots
        if datatype == 'Clock':
            minZ -= 1e-8
            maxZ += 1e-8
            
        logging.info("Autoset min: %f, max: %f" % (minZ, maxZ))
 
    # if user-defined number of col use that
    if NColFig != 0: Nc = NColFig
    else: Nc = int(np.ceil(np.sqrt(Nplots)))
    Nr = int(np.ceil(np.float(Nplots)/Nc))
    
    if figSize[0] == 0:
        if makeMovie: figSize[0]=5+2*Nc
        else: figSize[0]=10+3*Nc
    if figSize[1] == 0:
        if makeMovie: figSize[1]=4+1*Nr
        else: figSize[1]=8+2*Nr
 
    figgrid, axa = plt.subplots(Nr, Nc, sharex=True, sharey=True, figsize=figSize)
 
    if Nplots == 1: axa = np.array([axa])
    figgrid.subplots_adjust(hspace=0, wspace=0)
    axaiter = chain.from_iterable(axa)
 
    # axes label
    if len(axa.shape) == 1: # only one row
        [ax.set_xlabel(axesInPlot[0]+xlabelunit, fontsize=20) for ax in axa[:]]
        if cmesh:
            axa[0].set_ylabel(axesInPlot[1]+ylabelunit, fontsize=20)
        else:
            axa[0].set_ylabel(datatype+ylabelunit, fontsize=20)
    else:
        [ax.set_xlabel(axesInPlot[0]+xlabelunit, fontsize=20) for ax in axa[-1,:]]
        if cmesh:
            [ax.set_ylabel(axesInPlot[1]+ylabelunit, fontsize=20) for ax in axa[:,0]]
        else:
            [ax.set_ylabel(datatype+ylabelunit, fontsize=20) for ax in axa[:,0]]
 
    # if gaps in time, collapse and add a black vertical line on separation points
    if axesInPlot[0] == 'time' and cmesh == False:
        delta = np.abs(xvals[:-1] - xvals[1:])
        jumps = np.where( delta > 100*np.median(delta) )[0] # jump if larger than 100 times the minimum step
        # remove jumps
        for j in jumps: xvals[j+1:] -= delta[j]
        gap = xvals[-1] / 100 # 1%
        for j in jumps: xvals[j+1:] += gap
 
    im = None
    for Ntab, title in enumerate(titles):
 
        ax = axa.flatten()[Ntab]
        ax.text(.5, .9, title, horizontalalignment='center', fontsize=14, transform=ax.transAxes)
 
        # add vertical lines and numbers at jumps (numbers are the jump sizes)
        if axesInPlot[0] == 'time' and cmesh == False and not np.all(np.isnan(dataCube[Ntab].filled(np.nan))):
            flat = dataCube[Ntab].filled(np.nan).flatten()
            [ ax.axvline(xvals[j]+gap/2., color='k') for j in jumps ]
            if minZ != 0: texty = minZ + np.abs(np.nanmin(flat))*0.01
            else: texty = np.nanmin(dataCube[Ntab]) + np.abs(np.nanmin(flat))*0.01
            [ ax.text( xvals[j]+gap/2., texty, '%.0f' % delta[j], fontsize=10 ) for j in jumps ]
 
        # set log scales if activated
        if 'X' in log: ax.set_xscale('log')
        if 'Y' in log: ax.set_yscale('log')
 
        colors = cycle(['#377eb8','#b88637','#4daf4a','#984ea3','#ffff33','#f781bf'])
        for Ncol, data in enumerate(dataCube[Ntab]):
 
            # set color, use defined colors if a few lines, otherwise a continuum colormap
            if len(dataCube[Ntab]) <= 6:
                color = next(colors)
                colorFlag = '#e41a1c'
            else:
                color = plt.cm.jet(Ncol/float(len(dataCube[Ntab])-1)) # from 0 to 1
                colorFlag = 'k'
 
            vals = dataCube[Ntab][Ncol]
            if np.ma.getmask(dataCube[Ntab][Ncol]).all():
                continue
 
            # 3D cmesh plot
            if cmesh:
               # stratch the imshow output to fill the plot size
                bbox = ax.get_window_extent().transformed(figgrid.dpi_scale_trans.inverted())
                aspect = ((xvals[-1]-xvals[0])*bbox.height)/((yvals[-1]-yvals[0])*bbox.width)
                if 'Z' in log:
                    if minZ == 0: minZ = np.log10(1e-6)
                    else: minZ = np.log10(minZ)
                    maxZ = np.log10(maxZ)
                    vals = np.log10(vals)
 
                if datatype == 'phase' or datatype == 'rotation':
                    #cmap = phase_colormap
                    cmap = plt.cm.hsv
                else:
                    try:
                        cmap = plt.cm.viridis
                    except AttributeError:
                        cmap = plt.cm.rainbow

                if not np.all(np.diff(yvals) == np.diff(yvals)[0]): # if not evenly spaced
                    gapsize = np.amax(np.abs(np.diff(yvals)))
                    gapidx = np.argmax(np.abs(np.diff(yvals)))  # index BEFORE the gap
                    y_spacing = np.mean(np.diff(yvals)[np.diff(yvals) != gapsize])  # y-spacing outside of the gap
                    if gapsize > 2*y_spacing: # if gap greater 2 times y-spacing, pad NaNs
                        vals = np.insert(vals, gapidx+1, np.nan * np.ones((int(gapsize / y_spacing),len(vals[0]))), axis=0)  # fill gap with NaNs

                # ugly fix to enforce min/max as imshow has some problems with very large numbers
                if not np.isnan(vals).all():
                    with np.errstate(invalid='ignore'): # silence warnings that occure for comparison to NaN values
                        vals.data[vals.filled(np.nanmedian(vals.data)) > maxZ] = maxZ
                        vals.data[vals.filled(np.nanmedian(vals.data)) < minZ] = minZ

                im = ax.imshow(vals.filled(np.nan), origin='lower', interpolation="none", cmap=cmap, norm=None, \
                        extent=[xvals[0],xvals[-1],yvals[0],yvals[-1]], aspect=str(aspect), vmin=minZ, vmax=maxZ)
 
            # make an antenna plot
            elif antCoords != []:
                ax.set_xlabel('')
                ax.set_ylabel('')
                ax.axes.get_xaxis().set_ticks([])
                ax.axes.get_yaxis().set_ticks([])
                #vals = (vals-0.9)/(1.1-0.9)
                areas = ( 5+vals*10 )**2 # normalize marker diameter in pts**2 to 15-30 pt - assumes vals are between 0 and 1!
                ax.scatter(antCoords[0], antCoords[1], c=vals, s=areas, cmap=plt.cm.jet, vmin=-0.5, vmax=0.5)
                size = np.max( [np.max(antCoords[0])-np.min(antCoords[0]), np.max(antCoords[1])-np.min(antCoords[1])] )*1.1 # make img squared
                ax.set_xlim( xmin=np.median(antCoords[0])-size/2., xmax=np.median(antCoords[0])+size/2. )
                ax.set_ylim( ymin=np.median(antCoords[1])-size/2., ymax=np.median(antCoords[1])+size/2. )
 
            # 2D scatter plot
            else:
                ax.plot(xvals[~vals.mask], vals[~vals.mask], 'o', color=color, markersize=markerSize, markeredgecolor='none') # flagged data are automatically masked
                if plotFlag:
                    ax.plot(xvals[vals.mask], vals.data[vals.mask], 'o', color=colorFlag, markersize=markerSize, markeredgecolor='none') # plot flagged points
                ax.set_xlim(xmin=min(xvals), xmax=max(xvals))
                ax.set_ylim(ymin=minZ, ymax=maxZ)
 
    if not im is None:
        # add a color bar to show scale
        figgrid.colorbar(im, ax=axa.ravel().tolist(), use_gridspec=True, fraction=0.02, pad=0.005, aspect=35)
 
    logging.info("Saving "+filename+'.png')
    try:
        figgrid.savefig(filename+'.png', bbox_inches='tight')
    except:
        figgrid.tight_layout()
        figgrid.savefig(filename+'.png')

    del im, figgrid
    plt.close()


def run(soltab, axesInPlot, axisInTable='', axisInCol='', axisDiff='', NColFig=0, figSize=[0,0], markerSize=2, minmax=[0,0], log='', \
               plotFlag=False, doUnwrap=False, refAnt='', refDir='', soltabsToAdd='', makeAntPlot=False, makeMovie=False, prefix='', ncpu=0):
    """
    This operation for LoSoTo implements basic plotting
    WEIGHT: flag-only compliant, no need for weight

    Parameters
    ----------
    axesInPlot : array of str
        1- or 2-element array which says the coordinates to plot (2 for 3D plots).

    axisInTable : str, optional
        the axis to plot on a page - e.g. ant to get all antenna's on one file. By default ''.

    axisInCol : str, optional
        The axis to plot in different colours - e.g. pol to get correlations with different colors. By default ''.

    axisDiff : str, optional
        This must be a len=2 axis and the plot will have the differential value - e.g. 'pol' to plot XX-YY. By default ''.

    NColFig : int, optional
        Number of columns in a multi-table image. By default is automatically chosen.

    figSize : array of int, optional
        Size of the image [x,y], if one of the values is 0, then it is automatically chosen. By default automatic set.

    markerSize : int, optional
        Size of the markers in the 2D plot. By default 2.

    minmax : array of float, optional
        Min max value for the independent variable (0 means automatic). By default 0.

    log : bool, optional
        Use Log='XYZ' to set which axes to put in Log. By default ''.

    plotFlag : bool, optional
        Whether to plot also flags as red points in 2D plots. By default False.

    doUnwrap : bool, optional
        Unwrap phases. By default False.

    refAnt : str, optional
        Reference antenna for phases. By default None.

    refDir : str, optional
        Reference direction for phases. By default None.

    soltabsToAdd : str, optional
        Tables to "add" (e.g. 'sol000/tec000'), it works only for tec and clock to be added to phases. By default None.

    makeAntPlot : bool, optional
        Make a plot containing antenna coordinates in x,y and in color the value to plot, axesInPlot must be [ant]. By default False.

    makeMovie : bool, optional
        Make a movie summing up all the produced plots, by default False.

    prefix : str, optional
        Prefix to add before the self-generated filename, by default None.

    ncpu : int, optional
        Number of cpus, by default all available.
    """

    import os
    import numpy as np

    logging.info("Plotting soltab: "+soltab.name)

    # input check

    # str2list
    if axisInTable == '': axisInTable = []
    else: axisInTable = [axisInTable]
    if axisInCol == '': axisInCol = []
    else: axisInCol = [axisInCol]
    if axisDiff == '': axisDiff = []
    else: axisDiff = [axisDiff]

    if len(set(axisInTable+axesInPlot+axisInCol+axisDiff)) != len(axisInTable+axesInPlot+axisInCol+axisDiff):
        logging.error('Axis defined multiple times.')
        return 1

    # just because we use lists, check that they are 1-d
    if len(axisInTable) > 1 or len(axisInCol) > 1 or len(axisDiff) > 1:
        logging.error('Too many TableAxis/ColAxis/DiffAxis, they must be at most one each.')
        return 1

    for axis in axesInPlot+axisInCol+axisDiff:
        if axis not in soltab.getAxesNames():
            logging.error('Axis \"'+axis+'\" not found.')
            return 1

    if makeMovie:
        prefix = prefix+'__tmp__'

    if os.path.dirname(prefix) != '' and not os.path.exists(os.path.dirname(prefix)):
        logging.debug('Creating '+os.path.dirname(prefix)+'.')
        os.makedirs(os.path.dirname(prefix))

    if refAnt == '': refAnt = None
    elif refAnt != 'closest' and not refAnt in soltab.getAxisValues('ant', ignoreSelection = True):
        logging.warning('Reference antenna '+refAnt+' not found. Using: '+soltab.getAxisValues('ant')[1])
        refAnt = soltab.getAxisValues('ant')[1]

    if refDir == '': refDir = None
    elif refDir != 'center' and not refDir in soltab.getAxisValues('dir', ignoreSelection = True):
        logging.error('Reference direction '+refDir+' not found. Using: '+soltab.getAxisValues('dir')[1])
        refDir = soltab.getAxisValues('dir')[1]

    minZ, maxZ = minmax

    solset = soltab.getSolset()
    soltabsToAdd = [ solset.getSoltab(soltabName) for soltabName in soltabsToAdd ]

    cmesh = False
    if len(axesInPlot) == 2:
        cmesh = True
        # not color possible in 3D
        axisInCol = []
    elif len(axesInPlot) != 1:
        logging.error('Axes must be a len 1 or 2 array.')
        return 1
    # end input check

    # all axes that are not iterated by anything else
    axesInFile = soltab.getAxesNames()
    for axis in axisInTable+axesInPlot+axisInCol+axisDiff:
        axesInFile.remove(axis)

    # set subplots scheme
    if axisInTable != []:
        Nplots = soltab.getAxisLen(axisInTable[0])
    else:
        Nplots = 1

    # prepare antennas coord in makeAntPlot case
    if makeAntPlot:
        if axesInPlot != ['ant']:
            logging.error('If makeAntPlot is selected the "Axes" values must be "ant"')
            return 1
        antCoords = [[],[]]
        for ant in soltab.getAxisValues('ant'): # select only user-selected antenna in proper order
            antCoords[0].append(+1*soltab.getSolset().getAnt()[ant][1])
            antCoords[1].append(-1*soltab.getSolset().getAnt()[ant][0])

    else:
        antCoords = []

    datatype = soltab.getType()

    # start processes for multi-thread
    mpm = multiprocManager(ncpu, _plot)

    # compute dataCube size
    shape = []
    if axisInTable != []: shape.append(soltab.getAxisLen(axisInTable[0]))
    else: shape.append(1)
    if axisInCol != []: shape.append(soltab.getAxisLen(axisInCol[0]))
    else: shape.append(1)
    if cmesh:
        shape.append(soltab.getAxisLen(axesInPlot[1]))
        shape.append(soltab.getAxisLen(axesInPlot[0]))
    else:
        shape.append(soltab.getAxisLen(axesInPlot[0]))
    
    # will contain the data to pass to each thread to make 1 image
    dataCube = np.ma.zeros( shape=shape, fill_value=np.nan )

    # cycle on files
    if makeMovie: pngs = [] # store png filenames
    for vals, coord, selection in soltab.getValuesIter(returnAxes=axisDiff+axisInTable+axisInCol+axesInPlot):

        # set filename
        filename = ''
        for axis in axesInFile:
            filename += axis+str(coord[axis])+'_'
        filename = filename[:-1] # remove last _
        if prefix+filename == '': filename = 'plot'

        # axis vals (they are always the same, regulat arrays)
        xvals = coord[axesInPlot[0]]
        # if plotting antenna - convert to number
        if axesInPlot[0] == 'ant':
            xvals = np.arange(len(xvals))

        # if plotting time - convert in h/min/s
        xlabelunit=''
        if axesInPlot[0] == 'time':
            if xvals[-1] - xvals[0] > 3600:
                xvals = (xvals-xvals[0])/3600.  # hrs
                xlabelunit = ' [hr]'
            elif xvals[-1] - xvals[0] > 60:
                xvals = (xvals-xvals[0])/60.   # mins
                xlabelunit = ' [min]'
            else:
                xvals = (xvals-xvals[0])  # sec
                xlabelunit = ' [s]'
        # if plotting freq convert in MHz
        elif axesInPlot[0] == 'freq':
            xvals = xvals/1.e6 # MHz
            xlabelunit = ' [MHz]'

        if cmesh:
            # axis vals (they are always the same, regular arrays)
            yvals = coord[axesInPlot[1]]
            # same as above but for y-axis
            if axesInPlot[1] == 'ant':
                yvals = np.arange(len(yvals))

            if len(xvals) <= 1 or len(yvals) <=1:
                logging.error('3D plot must have more then one value per axes.')
                mpm.wait()
                return 1

            ylabelunit=''
            if axesInPlot[1] == 'time':
                if yvals[-1] - yvals[0] > 3600:
                    yvals = (yvals-yvals[0])/3600.  # hrs
                    ylabelunit = ' [hr]'
                elif yvals[-1] - yvals[0] > 60:
                    yvals = (yvals-yvals[0])/60.   # mins
                    ylabelunit = ' [min]'
                else:
                    yvals = (yvals-yvals[0])  # sec
                    ylabelunit = ' [s]'
            elif axesInPlot[1] == 'freq':  # Mhz
                yvals = yvals/1.e6
                ylabelunit = ' [MHz]'
        else:
            yvals = None
            if datatype == 'clock':
                datatype = 'Clock'
                ylabelunit = ' (s)'
            elif datatype == 'tec':
                datatype = 'dTEC'
                ylabelunit = ' (TECU)'
            elif datatype == 'rotationmeasure':
                datatype = 'dRM'
                ylabelunit = r' (rad m$^{-2}$)'
            elif datatype == 'tec3rd':
                datatype = r'dTEC$_3$'
                ylabelunit = r' (rad m$^{-3}$)'
            else:
                ylabelunit = ''

        # cycle on tables
        soltab1Selection = soltab.selection # save global selection and subselect only axex to iterate
        soltab.selection = selection
        titles = []

        for Ntab, (vals, coord, selection) in enumerate(soltab.getValuesIter(returnAxes=axisDiff+axisInCol+axesInPlot)):

            # set tile
            titles.append('')
            for axis in coord:
                if axis in axesInFile+axesInPlot+axisInCol: continue
                titles[Ntab] += axis+':'+str(coord[axis])+' '
            titles[Ntab] = titles[Ntab][:-1] # remove last ' '

            # cycle on colors
            soltab2Selection = soltab.selection
            soltab.selection = selection
            for Ncol, (vals, weight, coord, selection) in enumerate(soltab.getValuesIter(returnAxes=axisDiff+axesInPlot, weight=True, refAnt=refAnt, refDir=refDir)):

                # differential plot
                if axisDiff != []:
                    # find ordered list of axis
                    names = [axis for axis in soltab.getAxesNames() if axis in axisDiff+axesInPlot]
                    if axisDiff[0] not in names:
                        logging.error("Axis to differentiate (%s) not found." % axisDiff[0])
                        mpm.wait()
                        return 1
                    if len(coord[axisDiff[0]]) != 2:
                        logging.error("Axis to differentiate (%s) has too many values, only 2 is allowed." % axisDiff[0])
                        mpm.wait()
                        return 1

                    # find position of interesting axis
                    diff_idx = names.index(axisDiff[0])
                    # roll to first place
                    vals = np.rollaxis(vals,diff_idx,0)
                    vals = vals[0] - vals[1]
                    weight = np.rollaxis(weight,diff_idx,0)
                    weight[0][ weight[1]==0 ] = 0
                    weight = weight[0]
                    del coord[axisDiff[0]]

                # add tables if required (e.g. phase/tec)
                for soltabToAdd in soltabsToAdd:
                    logging.warning('soltabsToAdd not implemented. Ignoring.')
#                    newCoord = {}
#                    for axisName in coord.keys():
#                        # prepare selected on present axes
#                        if axisName in soltabToAdd.getAxesNames():
#                            if type(coord[axisName]) is np.ndarray:
#                                newCoord[axisName] = coord[axisName]
#                            else:
#                                newCoord[axisName] = [coord[axisName]] # avoid being interpreted as regexp, faster
#
#                    soltabToAdd.setSelection(**newCoord)
#                    valsAdd = np.squeeze(soltabToAdd.getValues(retAxesVals=False, weight=False, refAnt=refAnt))
#
#                    # add missing axes
#                    print ('shape:', vals.shape)
#                    for axisName in coord.keys():
#                        if not axisName in soltabToAdd.getAxesNames():
#                            # find axis positions
#                            axisPos = soltab.getAxesNames().index(axisName)
#                            # create a new axes for the table to add and duplicate the values
#                            valsAdd = np.expand_dims(valsAdd, axisPos)
#                            print ('shape to add:', valsAdd.shape)
#
#                    if soltabToAdd.getType() == 'clock':
#                        valsAdd = 2. * np.pi * valsAdd * coord['freq']
#                    elif soltabToAdd.getType() == 'tec':
#                        valsAdd = -8.44797245e9 * valsAdd / coord['freq']
#                    else:
#                        logging.warning('Only Clock or TEC can be added to solutions. Ignoring: '+soltabToAdd.getType()+'.')
#                        continue
#
#                    if valsAdd.shape != vals.shape:
#                        logging.error('Cannot combine the table '+soltabToAdd.getType()+' with '+soltab.getType()+'. Wrong shape.')
#                        mpm.wait()
#                        return 1
#
#                    vals += valsAdd

                # normalize
                if (soltab.getType() == 'phase' or soltab.getType() == 'scalarphase'):
                    vals = normalize_phase(vals)
                if (soltab.getType() == 'rotation'):
                    vals = np.mod(vals + np.pi/2., np.pi) - np.pi/2.

                # is user requested axis in an order that is different from h5parm, we need to transpose
                if cmesh:
                    if soltab.getAxesNames().index(axesInPlot[0]) < soltab.getAxesNames().index(axesInPlot[1]):
                        vals = vals.T
                        weight = weight.T

                # unwrap if required
                if (soltab.getType() == 'phase' or soltab.getType() == 'scalarphase') and doUnwrap:
                    if len(axesInPlot) == 1:
                        vals = unwrap(vals)
                    else:
                        flags = np.array((weight == 0), dtype=bool)
                        if not (flags == True).all():
                            vals = unwrap_2d(vals, flags, coord[axesInPlot[0]], coord[axesInPlot[1]])

                dataCube[Ntab,Ncol] = vals
                sel1 = np.where(weight == 0.)
                sel2 = np.where(np.isnan(vals))
                if cmesh:
                    dataCube[Ntab,Ncol,sel1[0],sel1[1]] = np.ma.masked
                    dataCube[Ntab,Ncol,sel2[0],sel2[1]] = np.ma.masked
                else:
                    dataCube[Ntab,Ncol,sel1[0]] = np.ma.masked
                    dataCube[Ntab,Ncol,sel2[0]] = np.ma.masked

            soltab.selection = soltab2Selection
            ### end cycle on colors

        # if dataCube too large (> 500 MB) do not go parallel
        if np.array(dataCube).nbytes > 1024*1024*500:
            logging.debug('Big plot, parallel not possible.')
            _plot(Nplots, NColFig, figSize, markerSize, cmesh, axesInPlot, axisInTable, xvals, yvals, xlabelunit, ylabelunit, datatype, prefix+filename, titles, log, dataCube, minZ, maxZ, plotFlag, makeMovie, antCoords, None)
        else:
            mpm.put([Nplots, NColFig, figSize, markerSize, cmesh, axesInPlot, axisInTable, xvals, yvals, xlabelunit, ylabelunit, datatype, prefix+filename, titles, log, np.ma.copy(dataCube), minZ, maxZ, plotFlag, makeMovie, antCoords]) # copy is necessary otherwise other cycles overwrite the dataCube

        if makeMovie: pngs.append(prefix+filename+'.png')

        soltab.selection = soltab1Selection
        ### end cycle on tables

    mpm.wait()
    del mpm

    if makeMovie:
        def long_substr(strings):
            """
            Find longest common substring
            """
            substr = ''
            if len(strings) > 1 and len(strings[0]) > 0:
                for i in range(len(strings[0])):
                    for j in range(len(strings[0])-i+1):
                        if j > len(substr) and all(strings[0][i:i+j] in x for x in strings):
                            substr = strings[0][i:i+j]
            return substr
        movieName = long_substr(pngs)
        assert movieName != '' # need a common prefix, use prefix keyword in case
        logging.info('Making movie: '+movieName)
        # make every movie last 20 sec, min one second per slide
        fps = np.ceil(len(pngs)/200.)
        ss="mencoder -ovc lavc -lavcopts vcodec=mpeg4:vpass=1:vbitrate=6160000:mbd=2:keyint=132:v4mv:vqmin=3:lumi_mask=0.07:dark_mask=0.2:"+\
                "mpeg_quant:scplx_mask=0.1:tcplx_mask=0.1:naq -mf type=png:fps="+str(fps)+" -nosound -o "+movieName.replace('__tmp__','')+".mpg mf://"+movieName+"*  > mencoder.log 2>&1"
        os.system(ss)
        #for png in pngs: os.system('rm '+png)

    return 0
